/*
 * Diskheap Library Main Routines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap.c,v 1.43 2002/03/26 15:34:33 gray Exp $
 */

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define __DISKHEAP_MAIN__

#include "conf.h"
#include "compat.h"
#include "diskheap.h"
#include "diskheap_free.h"
#include "diskheap_io.h"
#include "diskheap_os.h"
#include "diskheap_loc.h"
#include "diskheap_sub.h"

#if IDENT_WORKS
#ident "$Id: diskheap.c,v 1.43 2002/03/26 15:34:33 gray Exp $"
#else
static char *rcs_id = "$Id: diskheap.c,v 1.43 2002/03/26 15:34:33 gray Exp $";
#endif

/*
 * Version id for the library.  To change it, you also need to change
 * the files: NEWS, ChangeLog, RELEASE.html (in 2 places), and
 * diskheap.texi (in 2 places).
 */
#if IDENT_WORKS
#ident "$Diskheap Version 1.4.0 March 26, 2002 $"
#else
static char *version_id = "$Diskheap Version 1.4.0 March 26, 2002 $";
#endif

/*
 * We have the following static variable to support the seeding of the
 * random number generator.  There is a reentrance issue here but I
 * think it is minimal.  At the most this will be testing the
 * reentrance of srandom.
 */
static	int	srandom_called_b = 0;

/*
 * static void close_file
 *
 * DESCRIPTION:
 *
 * Close a diskheap structure
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer the diskheap structure we are freeing.
 */
static	void	close_file(diskheap_t *dh_p)
{
  if (dh_p->dh_fd >= 0) {
    _dh_os_close(dh_p->dh_fd);
    dh_p->dh_fd = -1;
  }
  if (dh_p->dh_buf != NULL) {
    free(dh_p->dh_buf);
    dh_p->dh_buf = NULL;
  }
  
  dh_p->dh_magic = 0;
  free(dh_p);
}

/**************************** retrieve functions *****************************/

/*
 * static void *retrieve
 *
 * DESCRIPTION:
 *
 * Read in a previously allocated block-number/offset/[size] from the
 * diskheap file.
 *
 * RETURNS:
 *
 * Success - If the buffer argument is specified and not NULL, it will
 * be returned.  If the buffer argument is NULL then this will return
 * a pointer to allocated memory holding the user's disk information
 * which must be passed to free to dealloc.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are retrieving.
 *
 * offset -> Offset inside of the block of the allocation we are
 * retrieving.
 *
 * buffer -> Buffer to store the data in or NULL to alloc.
 *
 * max_read_size -> Set the maximum number of bytes to read from the
 * data item or 0 to read all of the bytes.  If the buffer argument is
 * not NULL then this must be non-0.
 *
 * size_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the number of bytes in the total on-disk chunk, not
 * including any internal header information.
 *
 * type_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the type value which was passed to the store function.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
static	void	*retrieve(diskheap_t *dh_p, const unsigned int block_num,
			  const unsigned int offset,
			  void *buffer, const unsigned int max_read_size,
			  unsigned int *size_p, unsigned int *type_p,
			  int *error_p)
{
  block_head_t	head;
  void		*user_buf;
  unsigned int	read_size, left;
  int		ret, head_len;
  
  /* if our offset is > 0 then this should be a sub-block retrieve */
  if (offset > 0) {
    return _dh_read_sub_block(dh_p, block_num, offset, buffer, max_read_size,
			      NULL /* we don't care about the slot-size */,
			      size_p, type_p, error_p);
  }
  
  /*
   * We can either shift the stuff down or ask them to pass to a
   * diskheap_read_free call.  Alternatives would be to have a macro
   * in the diskheap.h to do the free which would shift the pointer by
   * some number of bytes, a call to return the size of the header
   * information, etc..  Also this function could return a pointer to
   * the user information _and_ a pointer to the pointer to free.
   */
  
  /* read in the first block */
  ret = _dh_read_first_block(dh_p, block_num, max_read_size, &head, &head_len);
  if (ret != DISKHEAP_ERROR_NONE) {
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* better be at the start of a user allocation */
  if (head.bh_block_type != BLOCK_TYPE_USER_ALL) {
    if (head.bh_block_type == BLOCK_TYPE_FREE) {
      SET_POINTER(error_p, DISKHEAP_ERROR_BLOCK_FREE);
      return NULL;
    }
    else {
      SET_POINTER(error_p, DISKHEAP_ERROR_LOCATION);
      return NULL;
    }
  }
  
  /* we should never have a user size of 0 */
  if (head.bh_user_size == 0) {
    SET_POINTER(error_p, DISKHEAP_ERROR_BLOCK_HEADER);
    return NULL;
  }
  
  /* determine how much of the user allocate we are going to read */
  read_size = head.bh_user_size;
  if (max_read_size > 0 && max_read_size < read_size) {
    read_size = max_read_size;
  }
  
  /* allocate our user buffer if necessary */
  if (buffer == NULL) {
    user_buf = malloc(read_size);
    if (user_buf == NULL) {
      SET_POINTER(error_p, DISKHEAP_ERROR_ALLOC);
      return NULL;
    }
  }
  else {
    user_buf = buffer;
  }
    
  /*
   * copy the user allocation information from the disk buffer into
   * the user buffer we will return
   */
  left = dh_p->dh_block_size - head_len;
  if (left > read_size) {
    left = read_size;
  }
  memmove(user_buf, dh_p->dh_buf + head_len, left);
  
  if (left < read_size) {
    /* read in the rest of the user allocation */
    ret = _dh_os_pread(dh_p, (char *)user_buf + left, read_size - left,
		       block_num + 1, 0 /* offset */);
    if (ret != DISKHEAP_ERROR_NONE) {
      if (user_buf != buffer) {
	free(user_buf);
      }
      SET_POINTER(error_p, ret);
      return NULL;
    }
  }
  
  SET_POINTER(size_p, head.bh_user_size);
  SET_POINTER(type_p, head.bh_user_type);
  SET_POINTER(error_p, DISKHEAP_ERROR_NONE);
  return user_buf;
}

/***************************** exported routines *****************************/

/*
 * diskheap_t *diskheap_create
 *
 * DESCRIPTION:
 *
 * Create a new diskheap file.
 *
 * RETURNS:
 *
 * Success - A valid diskheap structure which should be passed to
 * diskheap_close.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * file -> File to open for the diskheap.
 *
 * flags -> Diskheap flags which affect the characteristics of the
 * heap.
 *
 * block_size -> Set the block_size of the diskheap.  Must be a factor
 * of 512bytes.  Set to 0 (recommended) to take the default of 4096
 * bytes.
 *
 * heap_type -> A type value which will be passed back to the
 * diskheap_open call.  This allows you to specify the type of data
 * being stored in the heap.
 *
 * open_mode -> Mode bits to pass to the open system call.  Set to 0
 * to take the default which is 0644.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
diskheap_t	*diskheap_create(const char *file, const unsigned int flags,
				 const unsigned int block_size,
				 const unsigned int heap_type,
				 const unsigned int open_mode, int *error_p)
{
  unsigned int	b_size, buf_size, mode;
  diskheap_t	*dh_p;
  int		ret;
  
  if (file == NULL) {
    SET_POINTER(error_p, DISKHEAP_ERROR_ARG_NULL);
    return NULL;
  }
  if (BIT_IS_SET(flags, DISKHEAP_OPEN_READ_ONLY)) {
    SET_POINTER(error_p, DISKHEAP_ERROR_ARG_INVALID);
    return NULL;
  }
  if (block_size > 0) {
    if (block_size % BLOCK_SIZE_FACTOR != 0) {
      SET_POINTER(error_p, DISKHEAP_ERROR_BLOCK_SIZE);
      return NULL;
    }
    b_size = block_size;
  }
  else {
    b_size = DISKHEAP_DEFAULT_BLOCK_SIZE;
  }
  
  /* allocate the structure */
  dh_p = calloc(1, sizeof(*dh_p));
  if (dh_p == NULL) {
    SET_POINTER(error_p, DISKHEAP_ERROR_ALLOC);
    return NULL;
  }
  dh_p->dh_fd = -1;
  
  /* initialize the structure */
  dh_p->dh_magic = DISKHEAP_T_MAGIC;
  dh_p->dh_flags = flags;
  dh_p->dh_version = DISKHEAP_DEFAULT_VERSION;
  dh_p->dh_block_size = b_size;
  dh_p->dh_user_type = heap_type;
  
  /* handle the file's permissions */
  if (open_mode == 0) {
    mode = 0644;
  }
  else {
    mode = open_mode;
  }
  
  /* open the file */
  dh_p->dh_fd = _dh_os_open(file, 1 /* create-flag */, 0 /* read-write */,
			    mode);
  if (dh_p->dh_fd < 0) {
    close_file(dh_p);
    SET_POINTER(error_p, DISKHEAP_ERROR_OPEN);
    return NULL;
  }
  
  /*
   * setup our start of user allocations after the admin and free
   * storage in both blocks and offset.
   */
  dh_p->dh_min_block_n = (DISKHEAP_DEFAULT_USER_START + (b_size - 1)) / b_size;
  dh_p->dh_max_block_n = dh_p->dh_min_block_n;
  
  /*
   * Alloc a pair of block buffers for various transactions.  We
   * allocate 2 blocks so we can read in one with previous data in the
   * other and shift them down.
   */
  buf_size = b_size * 2;
  dh_p->dh_buf = malloc(buf_size);
  if (dh_p->dh_buf == NULL) {
    close_file(dh_p);
    SET_POINTER(error_p, DISKHEAP_ERROR_ALLOC);
    return NULL;
  }
  dh_p->dh_buf_bounds_p = dh_p->dh_buf + buf_size;
  
  /* init our sub-block stuff */
  ret = _dh_sub_init(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* start our random number generator if necessary */
  diskheap_seed_random(0);
  
  /* initialize our free list code */
  ret = _dh_free_startup(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* set our create and open times */
  dh_p->dh_create_time = _dh_os_time();
  dh_p->dh_open_time = dh_p->dh_create_time;
  
  /* create the header */
  ret = _dh_write_header(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /*
   * write the initial user block of 0s to extend the file so we don't
   * get read errors when we read in the free space area next time
   */
  ret = _dh_write_initial_user(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  BIT_SET(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  SET_POINTER(error_p, DISKHEAP_ERROR_NONE);
  return dh_p;
}

/*
 * int diskheap_set_free_space
 *
 * DESCRIPTION:
 *
 * Set the amount of disk space to reserve for free-space information.
 * The default is currently 1mb which should be able to store more
 * than 150,000 free slots in the file.  Each "free slot" represents a
 * block-number and size (in blocks) of a free area in the diskheap.
 * Contiguous free space is combined so each free area is bounded by
 * allocated areas.
 *
 * Free slots which cannot be accounted for in this area will not be
 * stored and may be lost however the default settings should make
 * this a rare occurrence.
 *
 * Note: this call has to be made _immediately_ following the call to
 * diskheap_create and no later.  Once the file has been created, it
 * cannot be adjusted.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 *
 * free_space -> Number of bytes of free space to allow at the front
 * of the diskheap file.  The minimum is currently 512 bytes although
 * going this low will almost guarantee lost free locations.  It must
 * be a multiple of 512.
 */
int	diskheap_set_free_space(diskheap_t *dh_p,
				const unsigned int free_space)
{
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (free_space < DISKHEAP_FREE_CHUNK) {
    return DISKHEAP_ERROR_ARG_INVALID;
  }
  if (free_space % 512 != 0) {
    return DISKHEAP_ERROR_ARG_INVALID;
  }
  if (! BIT_IS_SET(dh_p->dh_flags, DISKHEAP_JUST_CREATED)) {
    return DISKHEAP_ERROR_AFTER_CREATE;
  }
  
  /* set the start of the past the admin and free areas */
  dh_p->dh_min_block_n = (DISKHEAP_MAX_HEADER + free_space +
			  (dh_p->dh_block_size - 1)) / dh_p->dh_block_size;
  dh_p->dh_max_block_n = dh_p->dh_min_block_n;
  
  /*
   * write [another] initial user block of 0s to extend the file so we
   * don't get read errors when we read in the free space area next
   * time
   */
  return _dh_write_initial_user(dh_p);
}

/*
 * diskheap_t *diskheap_open
 *
 * DESCRIPTION:
 *
 * Open a diskheap file.
 *
 * RETURNS:
 *
 * Success - A valid diskheap structure which should be passed to
 * diskheap_close.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * file -> File to open for the diskheap.
 *
 * flags -> Diskheap flags which affect the characteristics of the
 * heap.
 *
 * heap_type_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set with the heap-type as given to diskheap_create.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
diskheap_t	*diskheap_open(const char *file, const unsigned int flags,
			       unsigned int *heap_type_p, int *error_p)
{
  diskheap_t	*dh_p;
  int		ret, buf_size;
  
  if (file == NULL) {
    SET_POINTER(error_p, DISKHEAP_ERROR_ARG_NULL);
    return NULL;
  }
  
  /* allocate the structure */
  dh_p = calloc(1, sizeof(*dh_p));
  if (dh_p == NULL) {
    SET_POINTER(error_p, DISKHEAP_ERROR_ALLOC);
    return NULL;
  }
  dh_p->dh_fd = -1;
  
  /* initialize the structure */
  dh_p->dh_magic = DISKHEAP_T_MAGIC;
  
  /* open the file */
  dh_p->dh_fd = _dh_os_open(file, 0 /* no-create */,
			    BIT_IS_SET(flags, DISKHEAP_OPEN_READ_ONLY),
			    0);
  if (dh_p->dh_fd < 0) {
    close_file(dh_p);
    SET_POINTER(error_p, DISKHEAP_ERROR_OPEN);
    return NULL;
  }
  
  /* now we read in the header */
  ret = _dh_read_header(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* our internal buffer size */
  buf_size = dh_p->dh_block_size * 2;
  
  /*
   * Alloc a pair of block buffers for various transactions.  We
   * allocate 2 blocks so we can read in one with previous data in the
   * other and shift them down.
   */
  dh_p->dh_buf = malloc(buf_size);
  if (dh_p->dh_buf == NULL) {
    close_file(dh_p);
    SET_POINTER(error_p, DISKHEAP_ERROR_ALLOC);
    return NULL;
  }
  dh_p->dh_buf_bounds_p = dh_p->dh_buf + buf_size;
  
  /* init our sub-block stuff */
  ret = _dh_sub_init(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* start our random number generator if necessary */
  diskheap_seed_random(0);
  
  /* initialize our free list code */
  ret = _dh_free_startup(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* set our open time */
  dh_p->dh_open_time = _dh_os_time();
  /*
   * if we are re-opening this heap at the same time we closed it then
   * adjust the old close back so we can indicate that an open and no
   * close has happened
   */
  if (dh_p->dh_close_time >= dh_p->dh_open_time) {
    dh_p->dh_close_time = dh_p->dh_open_time - 1;
  }
  
  /* we write our header right back again to set the open time */
  ret = _dh_write_header(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* read in the free chunks, if any */
  ret = _dh_read_free(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    close_file(dh_p);
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  SET_POINTER(heap_type_p, dh_p->dh_user_type);
  SET_POINTER(error_p, DISKHEAP_ERROR_NONE);
  return dh_p;
}

/*
 * int diskheap_set_sync_often
 *
 * DESCRIPTION:
 *
 * Set to how often to sync the main administrative header and the
 * free list to the disk.
 *
 * Note: this may cause a sync to happen immediately.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 *
 * sync_header -> How many store and delete transactions will cause
 * the admin header to be written to disk to update counters and other
 * stats.  Set to 0 to take the default which is every 100 trxn.  Set
 * to less than 0 for never in which case it will be updated when the
 * file is closed or when diskheap_sync_header() is called.
 *
 * sync_free -> How many store and delete transactions will cause the
 * free list to be written to disk to update which blocks are free and
 * which are in use.  Set to 0 to take the default which is every 100
 * trxn.  Set to less than 0 for never in which case it will be
 * updated when the file is closed or when diskheap_sync_free_list()
 * is called.
 */
int	diskheap_set_sync_often(diskheap_t *dh_p,
				const int sync_header,
				const int sync_free)
{
  int	ret;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  
  if (sync_header == 0) {
    dh_p->dh_sync_header = DEFAULT_SYNC_HEADER;
  }
  else if (sync_header < 0) {
    dh_p->dh_sync_header = 0;
    dh_p->dh_sync_header_c = 0;
  }
  else {
    dh_p->dh_sync_header = sync_header;
  }
  
  /* do we need to sync now */
  if (dh_p->dh_sync_header > 0
      && dh_p->dh_sync_header_c >= dh_p->dh_sync_header) {
    ret = _dh_write_header(dh_p);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    dh_p->dh_sync_header_c = 0;
  }
  
  if (sync_free == 0) {
    dh_p->dh_sync_free = DEFAULT_SYNC_FREE;
  }
  else if (sync_free < 0) {
    dh_p->dh_sync_free = 0;
    dh_p->dh_sync_free_list_c = 0;
  }
  else {
    dh_p->dh_sync_free = sync_free;
  }
  
  /* do we need to sync now */
  if (dh_p->dh_sync_free > 0
      && dh_p->dh_sync_free_list_c >= dh_p->dh_sync_free) {
    ret = _dh_write_free(dh_p);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    dh_p->dh_sync_free_list_c = 0;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_sync_header
 *
 * DESCRIPTION:
 *
 * Sync the header administrative information including counters,
 * transaction dates, and other statistics to disk.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 */
int	diskheap_sync_header(diskheap_t *dh_p)
{
  int	ret;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  
  /* sync the header */
  ret = _dh_write_header(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  dh_p->dh_sync_header_c = 0;
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_sync_free_list
 *
 * DESCRIPTION:
 *
 * Sync the free-list information to disk.  The free list manages
 * which blocks are free.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 */
int	diskheap_sync_free_list(diskheap_t *dh_p)
{
  int	ret;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  /* sync the header */
  ret = _dh_write_free(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  dh_p->dh_sync_free_list_c = 0;
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_close
 *
 * DESCRIPTION:
 *
 * Close a diskheap structure and associated file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to diskheap structure.
 */
int	diskheap_close(diskheap_t *dh_p)
{
  int	ret;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  /* set our close time */
  dh_p->dh_close_time = _dh_os_time();
  /*
   * if we are closing this heap at the same time we opened it then
   * adjust the close time ahead so we can indicate that an open and
   * close has happened
   */
  if (dh_p->dh_open_time >= dh_p->dh_close_time) {
    dh_p->dh_open_time = dh_p->dh_close_time - 1;
  }
  
  /*
   * we write our header to disk to set the close time and update the
   * times and stats
   */
  ret = _dh_write_header(dh_p);
  if (ret == DISKHEAP_ERROR_NONE) {
    ret = _dh_write_free(dh_p);
  }
  
  (void)_dh_free_shutdown(dh_p);
  
  close_file(dh_p);
  
  return ret;
}

/*
 * int diskheap_store
 *
 * DESCRIPTION:
 *
 * Store a buffer of bytes into the diskheap returning the
 * block-number and offset where it was written.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * buffer -> Buffer of anonymous memory to be written to disk.
 *
 * user_size -> Size of buffer in bytes.  Must be > 0.
 *
 * user_type -> Type number to be associated with the buffer that is
 * being stored.  This number will be returned when you retrieve the
 * block and can be used to verify that you are dealing with the right
 * kind of storage chunk.  Pass in 0 if none is desired.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number of the allocation.  Both the block-number and the offset are
 * needed to read the information later.  This argument cannot be
 * NULL.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the allocation.  Both the block-number and
 * the offset are needed to read the information later.  This argument
 * cannot be NULL.
 */
int	diskheap_store(diskheap_t *dh_p, const void *buffer,
		       const unsigned int user_size,
		       const unsigned int user_type,
		       unsigned int *block_num_p, unsigned int *offset_p)
{
  int		head_len, ret, end_b = 0;
  unsigned int	block_num, block_n, found_n;
  unsigned char	head_buf[BLOCK_MAX_HEAD_LEN];
  block_head_t	head;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (BIT_IS_SET(dh_p->dh_flags, DISKHEAP_OPEN_READ_ONLY)) {
    return DISKHEAP_ERROR_READ_ONLY;
  }
  if (buffer == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (user_size == 0) {
    return DISKHEAP_ERROR_ARG_INVALID;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  /* are we allocating a sub-block? */
  if (user_size <= dh_p->dh_sub_largest_size) {
    ret = _dh_sub_store(dh_p, buffer, user_size, user_type, block_num_p,
			offset_p);
    if (ret != DISKHEAP_ERROR_TOO_LARGE) {
      return ret;
    }
    /*
     * We make a guess to see if it can fit in a sub-block because we
     * don't know how many bytes the user-type is going to encode
     * into.  If we are wrong then we store it in its own block.
     */
  }
  
  /*
   * need to make a determination of split or whole blocks
   */
  
  /* set the fields in our new block */
  memset(&head, 0, sizeof(head));
  head.bh_block_type = BLOCK_TYPE_USER_ALL;
  head.bh_user_type = user_type;
  /* block-size handled below */
  head.bh_user_size = user_size;
  
  /*
   * initially we make a guess at the number of blocks with a 0 length
   * header which we may have to redo below
   */
  block_n = (0 + head.bh_user_size + dh_p->dh_block_size - 1) /
    dh_p->dh_block_size;
  do {
    head.bh_block_n = block_n;
    
    /* need to add the header onto the front of the first block */
    head_len = _dh_block_header(head_buf, head_buf + sizeof(head_buf), &head);
    if (head_len < 0) {
      return DISKHEAP_ERROR_INTERNAL;
    }
    
    /* calculate our block size */
    block_n = (head_len + user_size + dh_p->dh_block_size - 1) /
      dh_p->dh_block_size;
    
    /*
     * The problem here is that we are storing the number of blocks
     * for the user allocation in the header of the allocation.  As
     * the number of blocks increases, the chances that the storage of
     * the number of blocks might increase by a byte.  So we have to
     * do into a loop here until we get the block-size correct.
     */
    
  } while (head.bh_block_n < block_n);
  
  /*
   * We loop here until we get a free entry.  We may get a free slot
   * from the free list which really isn't free.
   */
  while (1) {
    
    /* see if we have any free blocks available */
    ret = _dh_free_get_blocks(dh_p, block_n, &block_num, &found_n);
    if (ret != DISKHEAP_ERROR_NONE) {
      break;
    }
    
    /* read it in to see if it is indeed free */
    ret = _dh_read_first_block(dh_p, block_num,
			       1 /* read minimal user bytes */, &head,
			       NULL /* head-length */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    /* do we have a valid free block? */
    if (head.bh_block_type == BLOCK_TYPE_FREE) {
      break;
    }
    
    /* the bad entry has already been removed and we go on */
    dh_p->dh_invalid_free_c++;
  }
  
  if (ret == DISKHEAP_ERROR_NO_FREE) {
    /* allocate space onto the end of the file */
    end_b = 1;
    block_num = dh_p->dh_max_block_n;
    found_n = block_n;
  }
  else if (ret != DISKHEAP_ERROR_NONE) {
    /* if there was some other error, then return it */
    return ret;
  }
  
  /*
   * Copy our header into the block buffer.  We have to do this
   * because the read above blows it away.
   */
  memmove(dh_p->dh_buf, head_buf, head_len);
  
  /* write out the header and the user data */
  ret = _dh_write_header_user(dh_p, head_len, buffer, user_size, block_num);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* adjust our stats */
  ret = _dh_account_alloc(dh_p, found_n, user_size, end_b);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  SET_POINTER(block_num_p, block_num);
  SET_POINTER(offset_p, 0);
  return DISKHEAP_ERROR_NONE;
}

/*
 * void *diskheap_retrieve
 *
 * DESCRIPTION:
 *
 * Read in a previously allocated block-number/offset/[size] from the
 * diskheap file.
 *
 * RETURNS:
 *
 * Success - Pointer to allocated memory holding the user's disk
 * information which must be passed to free to de-alloc.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are retrieving.
 *
 * offset -> Offset inside of the block of the allocation we are
 * retrieving.
 *
 * size_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the number of bytes in the total on-disk chunk, not
 * including any internal header information.
 *
 * type_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the type value which was passed to the store function.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
void	*diskheap_retrieve(diskheap_t *dh_p, const unsigned int block_num,
			   const unsigned int offset, unsigned int *size_p,
			   unsigned int *type_p, int *error_p)
{
  if (dh_p == NULL) {
    SET_POINTER(error_p, DISKHEAP_ERROR_ARG_NULL);
    return NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    SET_POINTER(error_p, DISKHEAP_ERROR_PNT);
    return NULL;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  return retrieve(dh_p, block_num, offset, NULL /* buffer */, 0 /* max read */,
		  size_p, type_p, error_p);
}

/*
 * int diskheap_retrieve_to_buf
 *
 * DESCRIPTION:
 *
 * Read in a previously allocated block-number/offset/[size] from the
 * diskheap file into the buffer provided.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are retrieving.
 *
 * offset -> Offset inside of the block of the allocation we are
 * retrieving.
 *
 * buffer -> Buffer to store the data in.
 *
 * max_read_size -> The maximum number of bytes to read from the data
 * item.  The buffer argument should be able to hold this number of
 * bytes.  Must be > 0.
 *
 * size_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the number of bytes in the total on-disk chunk, not
 * including any internal header information.
 *
 * type_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the type value which was passed to the store function.
 */
int	diskheap_retrieve_to_buf(diskheap_t *dh_p,
				 const unsigned int block_num,
				 const unsigned int offset, void *buffer,
				 const unsigned int max_read_size,
				 unsigned int *size_p, unsigned int *type_p)
{
  int	ret;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  /* if a buffer is specified then we must know its max-read-size */
  if (buffer == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (max_read_size == 0) {
    return DISKHEAP_ERROR_ARG_INVALID;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  /* this will return the buffer argument */
  (void)retrieve(dh_p, block_num, offset, buffer, max_read_size, size_p,
		 type_p, &ret);
  return ret;
}

/*
 * int diskheap_update
 *
 * DESCRIPTION:
 *
 * Update a previously allocated block with new data.  This calls
 * diskheap_delete and diskheap_store. If you wish to guarantee that a
 * program or system crash will not corrupt your data, you should set
 * the safer_b argument in which case it will do a diskheap_store
 * _first_, and then delete the old block.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * new_buffer -> Buffer of anonymous memory to be updated to disk.
 *
 * new_size -> Size of buffer in bytes.  Must be > 0.
 *
 * new_type -> Type number to be associated with the buffer that is
 * being written.  Pass in 0 if none is desired.
 *
 * old_block_num -> Block-number of the allocation we are updating.
 *
 * old_offset -> Offset inside of the block of the allocation we are
 * updating.
 *
 * safer_b -> Set to 1 to have this routine do the equivalent of a
 * diskheap_store _first_, and then the delete of the old block.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number of the allocation.  This argument cannot be NULL.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the allocation.  This argument cannot be
 * NULL.
 *
 * old_size_p <- Size of the user allocation that we updated.
 *
 * old_type_p <- Type of the user allocation that we updated.
 */
int	diskheap_update(diskheap_t *dh_p, const void *new_buffer,
			const unsigned int new_size,
			const unsigned int new_type,
			const unsigned int old_block_num,
			const unsigned int old_offset, const int safer_b,
			unsigned int *block_num_p, unsigned int *offset_p,
			unsigned int *old_size_p, unsigned int *old_type_p)
{
  int		ret;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (BIT_IS_SET(dh_p->dh_flags, DISKHEAP_OPEN_READ_ONLY)) {
    return DISKHEAP_ERROR_READ_ONLY;
  }
  if (new_buffer == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (new_size == 0) {
    return DISKHEAP_ERROR_ARG_INVALID;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  if (safer_b) {
    /* store the new block to disk */
    ret = diskheap_store(dh_p, new_buffer, new_size, new_type, block_num_p,
			 offset_p);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    /* now delete the old block once we know that the new one is okay */
    ret = diskheap_delete(dh_p, old_block_num, old_offset, old_size_p,
			  old_type_p);
  }
  else {
    /*
     * NOTE: we can do a lot better here by reading in the header,
     * looking to see if we have enough space already, then see if the
     * next block is free, or are we getting smaller, etc., etc..  But
     * the delete and store is an effective way to do this.  This does
     * do a possibly unnecessary write to the disk tho.
     */
    
    /* delete the old block first */
    ret = diskheap_delete(dh_p, old_block_num, old_offset, old_size_p,
			  old_type_p);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    /* now store the new block to disk */
    ret = diskheap_store(dh_p, new_buffer, new_size, new_type, block_num_p,
			 offset_p);
  }
  
  return ret;
}

/*
 * int diskheap_delete
 *
 * DESCRIPTION:
 *
 * Delete a previously stored block-number/offset section in the
 * diskheap file.  This puts it on the free list meaning that it can
 * be returned by the library at a later time.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are deleting.
 *
 * offset -> Offset inside of the block of the allocation we are
 * deleting.
 *
 * user_size_p <- Size of the user allocation in bytes.
 *
 * user_type_p <- Type number associated with the buffer by
 * diskheap_store.
 */
int	diskheap_delete(diskheap_t *dh_p, const unsigned int block_num,
			const unsigned int offset, unsigned int *user_size_p,
			unsigned int *user_type_p)
{
  block_head_t	head;
  int		ret, head_len;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (BIT_IS_SET(dh_p->dh_flags, DISKHEAP_OPEN_READ_ONLY)) {
    return DISKHEAP_ERROR_READ_ONLY;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  /* if the offset is > 0 then this should be a sub-block allocation */
  if (offset > 0) {
    return _dh_sub_delete(dh_p, block_num, offset, user_size_p, user_type_p);
  }
  
  /* read in the current block */
  ret = _dh_read_first_block(dh_p, block_num, 1 /* read minimal user bytes */,
			     &head, NULL);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* verify that it is a block allocation before we block it away */
  if (head.bh_block_type != BLOCK_TYPE_USER_ALL) {
    if (head.bh_block_type == BLOCK_TYPE_FREE) {
      return DISKHEAP_ERROR_BLOCK_FREE;
    }
    else {
      return DISKHEAP_ERROR_LOCATION;
    }
  }
  
  /* rewrite it with the free status */
  head.bh_block_type = BLOCK_TYPE_FREE;
  /* keep the same block-size and other user information */
  head_len = _dh_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p, &head);
  if (head_len < 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write out the header and no user data into the first block */
  ret = _dh_write_header_user(dh_p, head_len, NULL /* user buffer */,
			      0 /* user-size */, block_num);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* add the entry to the free list */
  ret = _dh_free_add_blocks(dh_p, block_num, head.bh_block_n);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* adjust our stats */
  ret = _dh_account_delete(dh_p, head.bh_block_n);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  SET_POINTER(user_size_p, head.bh_user_size);
  SET_POINTER(user_type_p, head.bh_user_type);
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_fsync
 *
 * DESCRIPTION:
 *
 * Call fsync on the file descriptor associated with the diskheap.  It
 * is designed to make sure that all buffered data gets moved to the
 * disk.  See the manual entry for fsync to determine what this does
 * in reality.
 *
 * Warning: even if the fsync succeeds, there is no guarantee that if
 * the system would immediately crash that the diskheap file would not
 * be corrupted.
 *
 * Warning: on some operating systems, fsync may not be available in
 * which case the DISKHEAP_ERROR_OS_CAPABLE error code will be
 * returned.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 */
int	diskheap_fsync(diskheap_t *dh_p)
{
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (BIT_IS_SET(dh_p->dh_flags, DISKHEAP_OPEN_READ_ONLY)) {
    return DISKHEAP_ERROR_READ_ONLY;
  }
  BIT_CLEAR(dh_p->dh_flags, DISKHEAP_JUST_CREATED);
  
  return _dh_os_fsync(dh_p->dh_fd);
}

/*
 * int diskheap_label_set
 *
 * DESCRIPTION:
 *
 * Associate a specific location with a string label in the diskheap
 * header.  This allows you to establish the start of an on disk
 * structure so the diskheap file can be self contained with no
 * outside reference information.
 *
 * Examples of usage include the starting point for a hash table, the
 * first entry in a list, or the administrative information for a
 * mini-file system.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * label -> Label we are associated with a position.  The string must
 * be less than or equal to 32 bytes in length or an error will be
 * returned.  If the label matches an already set entry, then it will
 * be overwritten if the overwrite flag is set to 1 otherwise an error
 * will be returned.
 *
 * block_num -> Block number we are associating with the label.  If
 * set to 0 then delete the associated label.
 *
 * offset -> Offset number we are associating with the label.
 *
 * overwrite_b -> If set to 1 then allow the new position to overwrite
 * a current one if the label matches an existing entry.
 */
int	diskheap_label_set(diskheap_t *dh_p, const char *label,
			   const unsigned int block_num,
			   const unsigned int offset,
			   const int overwrite_b)
{
  label_loc_t	*lab_p, *bounds_p, *empty_p = NULL;
  unsigned int	label_len;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (BIT_IS_SET(dh_p->dh_flags, DISKHEAP_OPEN_READ_ONLY)) {
    return DISKHEAP_ERROR_READ_ONLY;
  }
  if (label == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  label_len = strlen(label);
  if (label_len > sizeof(lab_p->ll_label)) {
    return DISKHEAP_ERROR_LABEL_LONG;
  }
  
  /* look for a matching label */
  bounds_p = dh_p->dh_labels + COUNT_OF(dh_p->dh_labels);
  for (lab_p = dh_p->dh_labels; lab_p < bounds_p; lab_p++) {
    if (lab_p->ll_length == label_len
	&& memcmp(lab_p->ll_label, label, label_len) == 0) {
      break;
    }
    if (empty_p == NULL && lab_p->ll_length == 0) {
      empty_p = lab_p;
    }
  }
  
  if (lab_p >= bounds_p) {
    
    /* no room at the inn? */
    if (empty_p == NULL) {
      return DISKHEAP_ERROR_LABELS_FULL;
    }
    lab_p = empty_p;
    
    /* copy the label into the slot including the \0 */
    memmove(lab_p->ll_label, label, label_len);
    lab_p->ll_length = label_len;
  }
  else {
    
    /* we found a match, overwrite? */
    if (! overwrite_b) {
      return DISKHEAP_ERROR_LABEL_EXISTS;
    }
    
    /* do we need to clear the entry? */
    if (block_num == 0) {
      lab_p->ll_length = 0;
    }
  }
  
  /* set new position */
  lab_p->ll_block_num = block_num;
  lab_p->ll_offset = offset;
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_label_get
 *
 * DESCRIPTION:
 *
 * Get the location associated with a string label in the diskheap
 * header.  See diskheap_label_set for more information.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * label -> Label we are looking up in the header.  The string must be
 * less than or equal to 32 bytes in length or an error will be
 * returned.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number associated with the label.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block associated with the label.
 */
int	diskheap_label_get(diskheap_t *dh_p, const char *label,
			   unsigned int *block_num_p,
			   unsigned int *offset_p)
{
  label_loc_t	*lab_p, *bounds_p;
  unsigned int	label_len;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (label == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  label_len = strlen(label);
  if (label_len > sizeof(lab_p->ll_label)) {
    return DISKHEAP_ERROR_LABEL_LONG;
  }
  if (block_num_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (offset_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  
  /* look for the label */
  bounds_p = dh_p->dh_labels + COUNT_OF(dh_p->dh_labels);
  for (lab_p = dh_p->dh_labels; lab_p < bounds_p; lab_p++) {
    if (lab_p->ll_length == label_len
	&& memcmp(lab_p->ll_label, label, label_len) == 0) {
      break;
    }
  }
  
  if (lab_p >= bounds_p) {
    return DISKHEAP_ERROR_LABEL_UNKNOWN;
  }
  
  SET_POINTER(block_num_p, lab_p->ll_block_num);
  SET_POINTER(offset_p, lab_p->ll_offset);
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_label_get_entry
 *
 * DESCRIPTION:
 *
 * There are a certain number (currently 10) of label and location
 * associations stored in the diskheap header.  This gets a specific
 * entry.  See diskheap_label_set for more information.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * entry_n -> Number of the label entry to get.
 *
 * label_p <- Pointer to a character pointer which will be set to an
 * allocated version of the label from the list.  This must be passed
 * to free to deallocate.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number associated with the label.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block associated with the label.
 */
int	diskheap_label_get_entry(diskheap_t *dh_p, const unsigned int entry_n,
				 char **label_p, unsigned int *block_num_p,
				 unsigned int *offset_p)
{
  label_loc_t	*lab_p;
  char		*str;
  
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (label_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (block_num_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (offset_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  
  /* is the entry past the end of the list? */
  if (entry_n >= COUNT_OF(dh_p->dh_labels)) {
    return DISKHEAP_ERROR_LABEL_BOUNDS;
  }
  lab_p = dh_p->dh_labels + entry_n;
  
  /* create a copy of the label string and pass it back */
  if (label_p != NULL) {
    str = malloc(lab_p->ll_length + 1);
    if (str == NULL) {
      return DISKHEAP_ERROR_ALLOC;
    }
    memmove(str, lab_p->ll_label, lab_p->ll_length);
    str[lab_p->ll_length] = '\0';
    *label_p = str;
  }
  
  SET_POINTER(block_num_p, lab_p->ll_block_num);
  SET_POINTER(offset_p, lab_p->ll_offset);
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_linear_first
 *
 * DESCRIPTION:
 *
 * Start the linear access function by setting the block-num and
 * offset to the first allocation in the heap.  If there are no
 * entries in the heap DISKHEAP_ERROR_NOT_FOUND will be returned.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * linear_p <-> Pointer to a linear access structure which will be
 * initialized to point to the first block in the heap.
 */
int	diskheap_linear_first(diskheap_t *dh_p, diskheap_linear_t *linear_p)
{
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (linear_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int diskheap_linear_next
 *
 * DESCRIPTION:
 *
 * Adjust the linear structure to reference the next allocation
 * location in the heap.  If there are no more entries in the heap
 * DISKHEAP_ERROR_NOT_FOUND will be returned.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * linear_p <-> Pointer to a linear access structure which will be
 * adjusted to point to the first block in the heap.
 */
int	diskheap_linear_next(diskheap_t *dh_p, diskheap_linear_t *linear_p)
{
  if (dh_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  if (dh_p->dh_magic != DISKHEAP_T_MAGIC) {
    return DISKHEAP_ERROR_PNT;
  }
  if (linear_p == NULL) {
    return DISKHEAP_ERROR_ARG_NULL;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * void diskheap_seed_random
 *
 * DESCRIPTION:
 *
 * This seeds the random number generator inside of the diskheap
 * library and sets a static flag which will not cause any more calls
 * to the random seed routine to be made.  This should be called
 * before any other diskheap calls are made to be effective.
 *
 * Note: this is usually only needed in testing since the library auto
 * seeds the random number generator the first time a diskheap is
 * opened or created.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * seed -> Seed for the random number generator.  0 for none.
 */
void	diskheap_seed_random(const int seed)
{
  /*
   * If we've not seeded the random number generator before, then do
   * so now.  There is a reentrance issue because srandom may be
   * called multiple times.
   */
  if (! srandom_called_b) {
    _dh_os_srandom(seed);
    srandom_called_b = 1;
  }
}

/*
 * const char *diskheap_strerror
 *
 * DESCRIPTION:
 *
 * Return string version of error codes.
 *
 * RETURNS:
 *
 * Returns the string equivalent of the error.
 *
 * ARGUMENTS:
 *
 * error -> the error value to lookup.
 */
const char	*diskheap_strerror(const int error)
{
  error_str_t	*err_p;
  
  for (err_p = errors; err_p->es_error != 0; err_p++) {
    if (err_p->es_error == error) {
      return err_p->es_string;
    }
  }
  
  return INVALID_ERROR;
}
