/*
 * Diskheap Library Free Link Management Routines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_free.c,v 1.22 2002/02/19 15:31:44 gray Exp $
 */

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define __DISKHEAP_MAIN__

#include "conf.h"
#include "compat.h"
#include "diskheap.h"
#include "diskheap_free.h"
#include "diskheap_io.h"
#include "diskheap_os.h"
#include "diskheap_loc.h"

#if IDENT_WORKS
#ident "$Id: diskheap_free.c,v 1.22 2002/02/19 15:31:44 gray Exp $"
#else
static char *rcs_id = "$Id: diskheap_free.c,v 1.22 2002/02/19 15:31:44 gray Exp $";
#endif

/*
 * static int random_level
 *
 * DESCRIPTION:
 *
 * Return a random level to be associated with a new free-list entry.
 *
 * RETURNS:
 *
 * Random level from 0 to max_level - 1.
 *
 * ARGUMENTS:
 *
 * max_level -> Maximum level of the free-list.
 */
static	int	random_level(const int max_level)
{
  int	level_c;
  
  for (level_c = 0; level_c < max_level; level_c++) {
    /*
     * Basically we count the number of times that the random number
     * generator returns an odd number in a row.  On average this
     * should return 0 1/2 the time, 1 1/4 of the time, 2 1/8 of a
     * time, and N 1/(2^(N - 1)) of the time.  This is what we want.
     * We could test for this in the configure scripts.
     *
     * Since many machines return random numbers which aren't that
     * random, there may be better ways of doing this.  In the past I
     * had (random() % 10000 >= 5000) or something but I'd rather not
     * have the % overhead here.
     */
    if (_dh_os_random() & 1) {
      break;
    }
  }
  
  return level_c;
}

/*
 * static skip_slot_t *find_slot
 *
 * DESCRIPTION:
 *
 * Look for a specific block-num and size pair in the free skip list.
 * If they exist then a pointer to the matching slot is returned
 * otherwise NULL.  Either way, the links that were traversed to get
 * there are set in the update slot which has the maximum number of
 * levels.
 *
 * RETURNS:
 *
 * Success - Pointer to the slot which matches the block-num and size
 * pair.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_list_b -> Sort by block-number otherwise sort by size.
 *
 * block_num -> Block number we are searching for.  This can be 0 if
 * we are searching for a size but don't care about where.
 *
 * size -> Size that we are searching for in bytes.
 */
static	skip_slot_t	*find_slot(diskheap_t *dh_p, const int block_list_b,
				   const unsigned int block_num,
				   const unsigned int size)
{
  int		level_c, cmp;
  skip_slot_t 	*slot_p, *found_p = NULL, *next_p, *update_p;
  
  level_c =  DISKHEAP_MAX_SKIP_LEVEL - 1;
  if (block_list_b) {
    slot_p = dh_p->dh_skip_block_p;
  }
  else {
    slot_p = dh_p->dh_skip_size_p;
  }
  
  /* get the update list */
  update_p = dh_p->dh_skip_update_p;
  
  /* traverse list to smallest entry */
  while (1) {
    
    /* next on we are looking for */
    next_p = slot_p->fs_next_p[level_c];
    
    /* are we are at the end of a row? */
    if (next_p == NULL) {
      cmp = 1;
    }
    else if (next_p == found_p) {
      /* did we find it again? */
      cmp = 0;
    }
    else if (block_list_b) {
      /* sort by block-num then size */
      cmp = next_p->fs_block_num - block_num;
      if (cmp == 0) {
	cmp = next_p->fs_size - size;
      }
    }
    else {
      /* sort by size then block-num */
      cmp = next_p->fs_size - size;
      /*
       * we only do our 2nd level sort if the size is the same and we
       * specified a block-num
       */
      if (cmp == 0 && block_num > 0) {
	cmp = next_p->fs_block_num - block_num;
      }
    }
    
    if (cmp < 0) {
      /* next slot is less, go right */
      slot_p = next_p;
      continue;
    }
    else if (cmp == 0) {
      /* we found a match but it may not be the first match */
      found_p = next_p;
    }
    
    /* we are lowering the level */
    
    update_p->fs_next_p[level_c] = slot_p;
    if (level_c == 0) {
      break;
    }
    level_c--;
  }
  
  return found_p;
}

/*
 * static int insert_entry
 *
 * DESCRIPTION:
 *
 * Insert an entry into a skip list.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_list_b -> Delete from the block-list otherwise from the
 * size-list.
 *
 * block_num -> Block-number of the free space.
 *
 * size -> Size, in blocks, of the free space.
 *
 * use_update_b -> Use the current pointers in dh_p->dh_skip_update_p
 * instead of calling find_slot.
 */
static	int	insert_entry(diskheap_t *dh_p, const int block_list_b,
			     const unsigned int block_num,
			     const unsigned int size, const int use_update_b)
{
  skip_slot_t	*new_p, *adjust_p, *update_p;
  int		level_n, level_c;
  
  /* find the slot in the list */
  if (use_update_b) {
    /* use the current pointers in dh_p->dh_skip_update_p */
  }
  else {
    if (find_slot(dh_p, block_list_b, block_num, size) != NULL) {
      /*
       * we should not have found it since that means that someone has
       * the same size and block-num
       */
      return DISKHEAP_ERROR_FREE_LIST;
    }
    /* find-slot updates the dh_p->dh_skip_update_p slot */
  }
  
  /* get the update list */
  update_p = dh_p->dh_skip_update_p;
  
  /* general the level for our new slot */
  level_n = random_level(DISKHEAP_MAX_SKIP_LEVEL);
  
  /* allocate a new free entry */
  new_p = calloc(1, SKIP_SLOT_SIZE(level_n + 1));
  if (new_p == NULL) {
    return DISKHEAP_ERROR_ALLOC;
  }
  
  /* fill it with our new data */
  new_p->fs_block_num = block_num;
  new_p->fs_size = size;
  
  /* update the block skip list */
  for (level_c = 0; level_c <= level_n; level_c++) {
    if (update_p->fs_next_p[level_c] == NULL) {
      new_p->fs_next_p[level_c] = NULL;
    }
    else {
      /*
       * We are inserting our new slot after each of the slots in the
       * update array.  So for each level, we get the slot we are
       * adjusting, we take it's next pointers and set them in the new
       * slot, and we point its next pointers to the new slot.
       */
      adjust_p = update_p->fs_next_p[level_c];
      new_p->fs_next_p[level_c] = adjust_p->fs_next_p[level_c];
      adjust_p->fs_next_p[level_c] = new_p;
    }
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * static void delete_entry
 *
 * DESCRIPTION:
 *
 * Find and delete the entry from the skip list.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_list_b -> Delete from the block-list otherwise from the
 * size-list.
 *
 * block_num -> Block number we are searching for.  This can be 0 if
 * we are searching for a size but don't care about where.
 *
 * size -> Size that we are searching for in bytes.
 *
 * closest_b -> If the closest_b argument is 1 then it will return the
 * closest match >= to this one otherwise it will find an exact match.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set with the block-number of the item we deleted.  We have
 * this here because we search for something in the size free list
 * when we don't know it's block-num but we need it.
 *
 * found_size_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set with the size in bytes of the item we deleted.  We have
 * this here because we search for something in the size free list
 * which is closest in size.
 */
static	int	delete_entry(diskheap_t *dh_p, const int block_list_b,
			     const unsigned int block_num,
			     const unsigned int size, const int closest_b,
			     unsigned int *block_num_p,
			     unsigned int *found_size_p)
{
  skip_slot_t	*adjust_p, *delete_p, *next_p, *update_p;
  int		level_c;
  
  /* get the update list */
  update_p = dh_p->dh_skip_update_p;
  
  /* find the slot in the list */
  delete_p = find_slot(dh_p, block_list_b, block_num, size);
  /* find-slot updates the dh_p->dh_skip_update_p slots */
  
  if (delete_p == NULL) {
    if (! closest_b) {
      /* we didn't find it so we can't remove it */
      return DISKHEAP_ERROR_NOT_FOUND;
    }
    
    /* chose the next block in the list which is the closest >= */
    next_p = update_p->fs_next_p[0];
    if (next_p == NULL) {
      /* there is no closest match */
      return DISKHEAP_ERROR_NOT_FOUND;
    }
    delete_p = next_p->fs_next_p[0];
    if (delete_p == NULL) {
      /* there is no closest match */
      return DISKHEAP_ERROR_NOT_FOUND;
    }
    
    /*
     * The question here is whether or not we should carve off any
     * remaining blocks and add them to the free list.  This will
     * lower internal fragmentation but raise external.
     *
     * Alternatives are "deferred" fragmentation in which case this
     * block gets put into some sort of list with a timestamp or
     * counter and after some point we decide that the block is not a
     * short-term block so we might as well split the block and take
     * the external fragmentation.
     */
  }
  
  /* update the block skip list */
  for (level_c = 0; level_c <= DISKHEAP_MAX_SKIP_LEVEL; level_c++) {
    
    /*
     * The update node holds pointers to the slots which are pointing
     * to the one we want since we need to update those pointers
     * ahead.
     */
    adjust_p = update_p->fs_next_p[level_c];
    
    /*
     * If the pointer in question is not pointing to the deleted slot
     * then the deleted slot is shorter than this level and we are
     * done.  This is guaranteed if we have a proper skip list.
     */
    if (adjust_p->fs_next_p[level_c] != delete_p) {
      break;
    }
    
    /*
     * We are deleting a slot after each of the slots in the update
     * array.  So for each level, we get the slot we are adjusting, we
     * set it's next pointers to the next pointers at the same level
     * from the deleted slot.
     */
    adjust_p->fs_next_p[level_c] = delete_p->fs_next_p[level_c];
  }
  
  /*
   * Sanity check here, we should always have at least 1 pointer to
   * the found node that we are deleting.
   */
  if (level_c == 0) {
    return DISKHEAP_ERROR_FREE_LIST;
  }
  
  SET_POINTER(block_num_p, delete_p->fs_block_num);
  SET_POINTER(found_size_p, delete_p->fs_size);
  
  /*
   * Remove the deleted slot.  We initially had a linked list of old
   * slots what we added this to but then I realized that I'd need to
   * garbage collect it once and a while in which case it became not
   * worth it.
   */
  free(delete_p);
  
  return DISKHEAP_ERROR_NONE;
}

/***************************** exported routines *****************************/

/*
 * int _dh_free_startup
 *
 * DESCRIPTION:
 *
 * Initialize the free slot structures in a new diskheap.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 */
int	_dh_free_startup(diskheap_t *dh_p)
{
  /* alloc the root slots in our block and size skip lists */
  dh_p->dh_skip_block_p = calloc(1, SKIP_SLOT_SIZE(DISKHEAP_MAX_SKIP_LEVEL));
  if (dh_p->dh_skip_block_p == NULL) {
    return DISKHEAP_ERROR_ALLOC;
  }
  dh_p->dh_skip_size_p = calloc(1, SKIP_SLOT_SIZE(DISKHEAP_MAX_SKIP_LEVEL));
  if (dh_p->dh_skip_size_p == NULL) {
    return DISKHEAP_ERROR_ALLOC;
  }
  /* alloc a slot which has pointers to slots of various levels */
  dh_p->dh_skip_update_p = calloc(1, SKIP_SLOT_SIZE(DISKHEAP_MAX_SKIP_LEVEL));
  if (dh_p->dh_skip_update_p == NULL) {
    return DISKHEAP_ERROR_ALLOC;
  }
  
  /* sanity check here */
  if (COUNT_OF(dh_p->dh_sub_sizes) != COUNT_OF(dh_p->dh_sub_free_p)) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_free_shutdown
 *
 * DESCRIPTION:
 *
 * Shutdown any structures in the free list.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 */
void	_dh_free_shutdown(diskheap_t *dh_p)
{
  skip_slot_t	*slot_p, *next_p;
  sub_slot_t	*sub_slot_p, *sub_next_p;
  unsigned int	slot_c;
  
  if (dh_p->dh_skip_block_p != NULL) {
    for (slot_p = dh_p->dh_skip_block_p->fs_next_p[0];
	 slot_p != NULL;
	 slot_p = next_p) {
      next_p = slot_p->fs_next_p[0];
      free(slot_p);
    }
    free(dh_p->dh_skip_block_p);
    dh_p->dh_skip_block_p = NULL;
  }
  if (dh_p->dh_skip_size_p != NULL) {
    for (slot_p = dh_p->dh_skip_size_p->fs_next_p[0];
	 slot_p != NULL;
	 slot_p = next_p) {
      next_p = slot_p->fs_next_p[0];
      free(slot_p);
    }
    free(dh_p->dh_skip_size_p);
    dh_p->dh_skip_size_p = NULL;
  }
  if (dh_p->dh_skip_update_p != NULL) {
    free(dh_p->dh_skip_update_p);
    dh_p->dh_skip_update_p = NULL;
  }
  
  /* now free our sub-block slots */
  for (slot_c = 0; slot_c < COUNT_OF(dh_p->dh_sub_free_p); slot_c++) {
    for (sub_slot_p = dh_p->dh_sub_free_p[slot_c];
	 sub_slot_p != NULL;
	 sub_slot_p = sub_next_p) {
      sub_next_p = sub_slot_p->ss_next_p;
      free(sub_slot_p);
    }
  }
}

/*
 * int _dh_free_add_blocks
 *
 * DESCRIPTION:
 *
 * Add a free entry to our administrative structures.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_num -> Block-number of the free space.
 *
 * size -> Size, in blocks, of the free space.
 */
int	_dh_free_add_blocks(diskheap_t *dh_p, const unsigned int block_num,
			    const unsigned int size)
{
  skip_slot_t	*prev_p, *next_p;
  unsigned int	next_block_num = 0, next_size = 0;
  int		ret, join_up_b;
  
  /*
   * We are looking for a specific slot but it better not match
   * otherwise we have 2 blocks in the same location of the same size
   * meaning an invalid free list or other screwup.  We are really
   * interested in have the update pointers filled out.
   */
  if (find_slot(dh_p, 1 /* sort-by-block */, block_num, size) != NULL) {
    /*
     * we can't have a complete match since that means that someone is
     * sitting in our chair at the same block-number!!!
     */
    return DISKHEAP_ERROR_FREE_LIST;
  }
  /* find_slot fills in dh_p->dh_skip_update_p */
  
  /*
   * Here is where we can take a look at the entries on either side of
   * us to see if either or both are contiguous so we can coalesce
   * them.  The update's lowest pointer is to the node directly to the
   * left of where we will be inserting.
   */
  prev_p = dh_p->dh_skip_update_p->fs_next_p[0];
  
  /*
   * next_p then points to the next entry in the list or the entry
   * after the one we are trying to insert
   */
  next_p = prev_p->fs_next_p[0];
  if (next_p != NULL && block_num + size == next_p->fs_block_num) {
    
    join_up_b = 1;
    
    /* record the block/size because we'll have to delete the next_p entry */
    next_block_num = next_p->fs_block_num;
    next_size = next_p->fs_size;
    
    /*
     * if we are joining with the next block, we have to remove it
     * from both the block-list as well as the size list.
     */
    ret = delete_entry(dh_p, 1 /* sort-by-block */, next_block_num,
		       next_size, 0 /* exact-match */,
		       NULL /* block-num-p */, NULL /* found-size-p */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    /* delete the next entry from the size skip-list */
    ret = delete_entry(dh_p, 0 /* sort-by-size */, next_block_num, next_size,
		       0 /* exact-match */, NULL /* block-num-p */,
		       NULL /* found-size-p */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
  }
  else {
    join_up_b = 0;
  }
  
  /* is the previous one contiguous */
  if (prev_p->fs_block_num + prev_p->fs_size == block_num) {
    
    /* previous block is contiguous */
    
    /*
     * we need to delete it from the size list and then re-add it
     * since it's size is changing so it might be in a different place
     * in the list
     */
    ret = delete_entry(dh_p, 0 /* sort-by-size */, prev_p->fs_block_num,
		       prev_p->fs_size, 0 /* exact-match */,
		       NULL /* block-num-p */, NULL /* found-size-p */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    /* if we are also combining upwards, then add it that size too */
    if (join_up_b) {
      /* just add in the next block's size to the previous */
      prev_p->fs_size += next_size;
    }
    
    /*
     * Add in our new size to the previous entry in the block-list.
     * We don't have to remove and then re-add it to the list since it
     * is at the same position.
     */
    prev_p->fs_size += size;
    
    /* insert the combined block with the new size into the size list */
    return insert_entry(dh_p, 0 /* sort-by-size */, prev_p->fs_block_num,
			prev_p->fs_size, 0 /* need to get update pointers */);
  }
  
  /* we know we only have a comb-up here */
  if (join_up_b) {
    
    /* next block is contiguous and it is already removed from the lists */
    
    /*
     * The new combined block-num is the lower one and we add in the
     * new blocks into the size.
     */
    next_block_num = block_num;
    next_size += size;
    
    /* insert the combined block with the new size into the block list */
    ret = insert_entry(dh_p, 1 /* sort-by-block */, next_block_num, next_size,
		       0 /* need to get update pointers */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    /* insert the combined block with the new size into the size list */
    return insert_entry(dh_p, 0 /* sort-by-size */, next_block_num, next_size,
			0 /* need to get update pointers */);
  }
  
  /*
   * Previous nor the next blocks are contiguous so just insert it
   * into the list.
   */
  ret = insert_entry(dh_p, 1 /* sort-by-block */, block_num, size,
		     1 /* use current update pointers */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* insert it into the size skip list */
  ret = insert_entry(dh_p, 0 /* sort-by-size */, block_num, size,
		     0 /* need to get update pointers */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_free_get_blocks
 *
 * DESCRIPTION:
 *
 * Find in the free list the best matching space to fit X blocks.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP_ERROR_NO_FREE is no [appropriate] free space is
 * available or another error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer to modify.
 *
 * size -> Number of free blocks we need for our allocation.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the block-number of the free block to use.
 *
 * found_size_np <- Pointer to an unsigned int which be set to the
 * number of blocks that we found in the free list.
 */
int	_dh_free_get_blocks(diskheap_t *dh_p, const unsigned int size,
			    unsigned int *block_num_p,
			    unsigned int *found_size_np)
{
  int		ret;
  unsigned int	block_num, found_size;
  
  /* try to find and delete an entry from the size skip list */ 
  ret = delete_entry(dh_p, 0 /* sort-by-size */, 0 /* block-number */,
		     size, 1 /* closest-match */, &block_num, &found_size);
  if (ret == DISKHEAP_ERROR_NONE) {
    
    /* delete it from the block-number skip-list */
    ret = delete_entry(dh_p, 1 /* sort-by-block */, block_num, found_size,
		       0 /* exact-match */, NULL /* block-num-p */,
		       NULL /* found-size-p */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    SET_POINTER(block_num_p, block_num);
    SET_POINTER(found_size_np, found_size);
    return DISKHEAP_ERROR_NONE;
  }
  else if (ret != DISKHEAP_ERROR_NOT_FOUND) {
    return ret;
  }
  
  /*
   * if there is a space free which is 1 block less than the needed
   * space then take it and get a 1 block free space and write the 1st
   * block there with a continuation pointer to the next block since
   * we have a high likely hood that we will have to do 2 reads
   * anyway.
   */
  
  /*
   * if there is space free but larger then consider splitting
   * (rules?)
   */
  
  /*
   * if there is space free but smaller then consider fragmenting
   * (rules?)
   */
  
  /*
   * maybe have a switch for speed versus size.  maybe have a maximum
   * heap size switch (for file system mode) were it would switch from
   * space to speed when the amount of free blocks get down below some
   * %.
   */
  
  return DISKHEAP_ERROR_NO_FREE;
}

/*
 * int _dh_free_get_one
 *
 * DESCRIPTION:
 *
 * Find in the free list the first free block in the heap and return
 * it.  This is usually used in the sub-block allocate when we need a
 * new block of a certain size.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP_ERROR_NO_FREE is no [appropriate] free space is
 * available or another error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer to modify.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the block-number of the free block to use.
 *
 * found_size_np <- Pointer to an unsigned int which be set to the
 * number of blocks that we found in the free list.
 */
int	_dh_free_get_one(diskheap_t *dh_p, unsigned int *block_num_p,
			 unsigned int *found_size_np)
{
  int		ret;
  unsigned int	block_num, found_size;
  
  /*
   * try to find and delete an entry from the block skip list so we
   * get the very first skip block in the heap.
   */ 
  ret = delete_entry(dh_p, 1 /* sort-by-block */, 0 /* block-number */,
		     1 /* we need 1 block */, 1 /* closest-match */,
		     &block_num, &found_size);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (ret == DISKHEAP_ERROR_NOT_FOUND) {
      return DISKHEAP_ERROR_NO_FREE;
    }
    else {
      return ret;
    }
  }
  
  /* delete it from the size skip-list */
  ret = delete_entry(dh_p, 0 /* sort-by-size */, block_num, found_size,
		     0 /* exact-match */, NULL /* block-num-p */,
		     NULL /* found-size-p */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  SET_POINTER(block_num_p, block_num);
  SET_POINTER(found_size_np, found_size);
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_free_add_bytes
 *
 * DESCRIPTION:
 *
 * Add a sub-block free entry to our administrative structures.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_num -> Block-number of the free space.
 *
 * offset -> Location in the block the free space.
 *
 * bytes_n -> Size of the free space.
 */
int	_dh_free_add_bytes(diskheap_t *dh_p, const unsigned int block_num,
			   const unsigned int offset,
			   const unsigned int bytes_n)
{
  sub_slot_t	*slot_p;
  unsigned int	size_c;
  
  /* find the matching size in our list of sub-blocks */
  for (size_c = 0; size_c < COUNT_OF(dh_p->dh_sub_free_p); size_c++) {
    if (dh_p->dh_sub_sizes[size_c] >= bytes_n) {
      break;
    }
  }
  
  /* we should have matched one of the slots */
  if (size_c >= COUNT_OF(dh_p->dh_sub_free_p)) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* allocate our free slot */
  slot_p = calloc(1, sizeof(*slot_p));
  if (slot_p == NULL) {
    return DISKHEAP_ERROR_ALLOC;
  }
  
  /* initialize the new slot */
  slot_p->ss_block_num = block_num;
  slot_p->ss_offset = offset;
  slot_p->ss_next_p = dh_p->dh_sub_free_p[size_c];
  
  /* add it into our linked lists */
  dh_p->dh_sub_free_p[size_c] = slot_p;
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_free_get_bytes
 *
 * DESCRIPTION:
 *
 * Find in the free list the best matching space to fit X bytes for
 * the sub-block allocations.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP_ERROR_NO_FREE is no [appropriate] free space is
 * available or another error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer to modify.
 *
 * bytes_n -> Number of free bytes we need for our allocation.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the block-number of the free block to use.
 *
 * offset_p <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the available space.
 *
 * slot_size_p <- Pointer to an unsigned int which be set to the size
 * of the slot that we were allocated.  This will be set if an
 * available free slot is found or not.
 */
int	_dh_free_get_bytes(diskheap_t *dh_p, const unsigned int bytes_n,
			   unsigned int *block_num_p, unsigned int *offset_p,
			   unsigned int *slot_size_p)
{
  sub_slot_t	*slot_p;
  unsigned int	size_c;
  
  /* find the matching size in our list of sub-blocks */
  for (size_c = 0; size_c < COUNT_OF(dh_p->dh_sub_free_p); size_c++) {
    if (dh_p->dh_sub_sizes[size_c] >= bytes_n) {
      break;
    }
  }
  
  /*
   * NOTE: we rely on this being set whether there are free slots
   * available or not
   */
  SET_POINTER(slot_size_p, dh_p->dh_sub_sizes[size_c]);
  
  /* we should have matched one of the slots */
  if (size_c >= COUNT_OF(dh_p->dh_sub_free_p)) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* no matching entries? */
  slot_p = dh_p->dh_sub_free_p[size_c];
  if (slot_p == NULL) {
    return DISKHEAP_ERROR_NO_FREE;
  }
  
  /* take it out of the free linked list */
  dh_p->dh_sub_free_p[size_c] = slot_p->ss_next_p;
  
  SET_POINTER(block_num_p, slot_p->ss_block_num);
  SET_POINTER(offset_p, slot_p->ss_offset);
  
  /* free up the slot */
  free(slot_p);
  
  return DISKHEAP_ERROR_NONE;
}
