/*
 * Diskheap FREE defines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_free.h,v 1.13 2002/02/19 15:31:45 gray Exp $
 */

#ifndef __DISKHEAP_FREE_H__
#define __DISKHEAP_FREE_H__

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

/*
 * This structure holds free-slot information in a skip-list format.
 * Skip lists have variable number of next pointers so the fs_next_p
 * pointer array may actually have more elements depending on how
 * large an allocation we make.  Don't you love these C hacks.
 */
typedef struct skip_slot_st {
  unsigned int		fs_block_num;	/* block number of free space */
  unsigned int		fs_size;	/* size of free space in blocks */
  /*
   * Array of next pointers.  This may extend past the end of the
   * function if we allocate for space larger than the structure.
   */
  struct skip_slot_st	*fs_next_p[1];
} skip_slot_t;

/*
 * This structure holds free space information for the sub-block
 * slots.  Since the sub-blocks are put into bucket of specific size,
 * we can represent that as a simple linked list and don't have to get
 * fancy.  In addition, since we [currently] can't reclaim sub-blocks,
 * there is no need to have anything sorted by block-number.
 */
typedef struct sub_slot_st {
  unsigned int		ss_block_num;	/* block number of free space */
  unsigned int		ss_offset;	/* offset in block of free space */
  /* no need for the size since the bucket number will give that to us */
  struct sub_slot_st	*ss_next_p;	/* pointer to the next free entry */
} sub_slot_t;

/*
 * This macro helps us determine how much memory we need to store to
 * hold all of the next pointers in the skip-list entry.
 */
#define SKIP_SLOT_SIZE(next_n)	\
	(sizeof(skip_slot_t) + sizeof(struct skip_slot_st *) * ((next_n) - 1))

/*<<<<<<<<<<  The below prototypes are auto-generated by fillproto */

/*
 * int _dh_free_startup
 *
 * DESCRIPTION:
 *
 * Initialize the free slot structures in a new diskheap.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 */
extern
int	_dh_free_startup(diskheap_t *dh_p);

/*
 * int _dh_free_shutdown
 *
 * DESCRIPTION:
 *
 * Shutdown any structures in the free list.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 */
extern
void	_dh_free_shutdown(diskheap_t *dh_p);

/*
 * int _dh_free_add_blocks
 *
 * DESCRIPTION:
 *
 * Add a free entry to our administrative structures.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_num -> Block-number of the free space.
 *
 * size -> Size, in blocks, of the free space.
 */
extern
int	_dh_free_add_blocks(diskheap_t *dh_p, const unsigned int block_num,
			    const unsigned int size);

/*
 * int _dh_free_get_blocks
 *
 * DESCRIPTION:
 *
 * Find in the free list the best matching space to fit X blocks.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP_ERROR_NO_FREE is no [appropriate] free space is
 * available or another error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer to modify.
 *
 * size -> Number of free blocks we need for our allocation.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the block-number of the free block to use.
 *
 * found_size_np <- Pointer to an unsigned int which be set to the
 * number of blocks that we found in the free list.
 */
extern
int	_dh_free_get_blocks(diskheap_t *dh_p, const unsigned int size,
			    unsigned int *block_num_p,
			    unsigned int *found_size_np);

/*
 * int _dh_free_get_one
 *
 * DESCRIPTION:
 *
 * Find in the free list the first free block in the heap and return
 * it.  This is usually used in the sub-block allocate when we need a
 * new block of a certain size.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP_ERROR_NO_FREE is no [appropriate] free space is
 * available or another error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer to modify.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the block-number of the free block to use.
 *
 * found_size_np <- Pointer to an unsigned int which be set to the
 * number of blocks that we found in the free list.
 */
extern
int	_dh_free_get_one(diskheap_t *dh_p, unsigned int *block_num_p,
			 unsigned int *found_size_np);

/*
 * int _dh_free_add_bytes
 *
 * DESCRIPTION:
 *
 * Add a sub-block free entry to our administrative structures.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk head we are working with.
 *
 * block_num -> Block-number of the free space.
 *
 * offset -> Location in the block the free space.
 *
 * bytes_n -> Size of the free space.
 */
extern
int	_dh_free_add_bytes(diskheap_t *dh_p, const unsigned int block_num,
			   const unsigned int offset,
			   const unsigned int bytes_n);

/*
 * int _dh_free_get_bytes
 *
 * DESCRIPTION:
 *
 * Find in the free list the best matching space to fit X bytes for
 * the sub-block allocations.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP_ERROR_NO_FREE is no [appropriate] free space is
 * available or another error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer to modify.
 *
 * bytes_n -> Number of free bytes we need for our allocation.
 *
 * block_num_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the block-number of the free block to use.
 *
 * offset_p <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the available space.
 *
 * slot_size_p <- Pointer to an unsigned int which be set to the size
 * of the slot that we were allocated.  This will be set if an
 * available free slot is found or not.
 */
extern
int	_dh_free_get_bytes(diskheap_t *dh_p, const unsigned int bytes_n,
			   unsigned int *block_num_p, unsigned int *offset_p,
			   unsigned int *slot_size_p);

/*<<<<<<<<<<   This is end of the auto-generated output from fillproto. */

#ifdef __cplusplus
}
#endif


#endif /* ! __DISKHEAP_FREE_H__ */
