/*
 * Diskheap Library IO Routines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_io.c,v 1.21 2002/02/19 15:32:33 gray Exp $
 */

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define __DISKHEAP_MAIN__

#include "conf.h"
#include "compat.h"
#include "diskheap.h"
#include "diskheap_free.h"
#include "diskheap_io.h"
#include "diskheap_os.h"
#include "diskheap_loc.h"

#if IDENT_WORKS
#ident "$Id: diskheap_io.c,v 1.21 2002/02/19 15:32:33 gray Exp $"
#else
static char *rcs_id = "$Id: diskheap_io.c,v 1.21 2002/02/19 15:32:33 gray Exp $";
#endif

/*
 * static int put_buf
 *
 * DESCRIPTION:
 *
 * Add to a buffer a bunch of bytes and advance the buffer pointer so
 * it is always at the end of the data.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp - Pointer to a buffer pointer.  We will be adding the data
 * items to this location and then advancing the pointer past the new
 * data.
 *
 * bounds_p - Boundary location past the end of the buffer.  It will
 * add the data items to the buffer until it reaches this pointer.  If
 * it goes to write to this address it will return -1.
 *
 * buf -> Buffer we are putting.
 *
 * len -> Length of the buffer we are putting.
 */
static	int	put_buf(unsigned char **buf_pp, const void *bounds_p,
			const unsigned char *buf, const int len)
{
  unsigned char	*buf_p = *buf_pp, *start_p = *buf_pp;
  
  /* are we about to overflow the buffer? */
  if (buf_p + len > (unsigned char *)bounds_p) {
    return -1;
  }
    
  memmove(buf_p, buf, len);
  buf_p += len;
  
  /* advance our pointer */
  *buf_pp = buf_p;
  
  return buf_p - (unsigned char *)start_p;
}

/*
 * static int get_buf
 *
 * DESCRIPTION:
 *
 * Get from a buffer a block of data of a certain length, advancing
 * the buffer pointer so it is always at the end of the data.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp <-> Pointer to our buffer which we will be adding
 * information to and advancing.
 *
 * bounds_p -> Boundary location past the end of the buffer.  It will
 * get the data items from the buffer until it reaches this pointer.
 * If it goes to access this pointer, it will return -1.
 *
 * buf <-> Pointer to a buffer that we are writing into.
 *
 * len -> Length of the buffer we are copying into.
 */
static	int	get_buf(unsigned char **buf_pp, const void *bounds_p,
			unsigned char *buf, const int len)
{
  unsigned char	*buf_p = *buf_pp, *start_p = *buf_pp;
  
  /* are we about to overflow the buffer? */
  if (buf_p + len > (unsigned char *)bounds_p) {
    return -1;
  }
  
  memmove(buf, buf_p, len);
  buf_p += len;
  
  /* advance our pointer */
  *buf_pp = buf_p;
  
  return buf_p - (unsigned char *)start_p;
}

/*
 * static int put_uint
 *
 * DESCRIPTION:
 *
 * Add to a buffer a packed decimal value, advancing the buffer
 * pointer so it is always at the end of the data.  It will pack the
 * decimal into as few bytes as possible so get_uint must be
 * used on the other side to unpack.
 *
 * RETURNS:
 *
 * Success - Number of characters added to the string buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp <-> Pointer to our buffer which we will be adding
 * information to and advancing.
 *
 * bounds_p -> Boundary location past the end of the buffer.  It will
 * add the data items to the buffer until it reaches this pointer.  If
 * it goes to write to this address it will return -1.
 *
 * val -> Unsigned integer value that we are packing into the buffer.
 */
static	int	put_uint(unsigned char **buf_pp, const void *bounds_p,
			 const unsigned int val)
{
  unsigned int	work = val;
  unsigned char	ch, *buf_p = *buf_pp;
  unsigned int	ch_c;
  
  for (ch_c = 1;; ch_c++) {
    ch = work & 0177;
    work >>= 7;
    
    /* turn on high bit if there is more */
    if (work > 0) {
      ch |= 0200;
    }
    
    if (buf_p >= (unsigned char *)bounds_p) {
      return -1;
    }
    
    *buf_p++ = ch;
    
    /* we break here because we want 0 to be reflected by '\0' */
    if (work == 0) {
      break;
    }
  }
  
  /* advance our pointer */
  *buf_pp = buf_p;
  
  return ch_c;
}

/*
 * static int get_uint
 *
 * DESCRIPTION:
 *
 * Get from the buffer a packed decimal value, advancing the buffer
 * pointer so it is always at the end of the data.  The value must
 * have been stored with the put_uint.
 *
 * RETURNS:
 *
 * Success - Number of characters taken from the buffer.
 *
 * Failure - -1 if the maximum buffer pointer was reached.
 *
 * ARGUMENTS:
 *
 * buf_pp <-> Pointer to our buffer which we will be unpacking the
 * value from and advancing.
 *
 * bounds_p -> Boundary location past the end of the buffer.  It will
 * get the data items from the buffer until it reaches this pointer.
 * If it goes to access this pointer, it will return -1.
 *
 * val_p <- Pointer to an unsigned integer which will be assigned the
 * unpacked value.
 */
static	int	get_uint(unsigned char **buf_pp, const void *bounds_p,
			 unsigned int *val_p)
{
  unsigned int	answer = 0, work = 0, shift = 0;
  unsigned char	ch, *buf_p = *buf_pp;
  unsigned int	ch_c;
  
  for (ch_c = 1;; ch_c++) {
    if (buf_p >= (unsigned char *)bounds_p) {
      return -1;
    }
    ch = *buf_p++;
    work = ch & 0177;
    
    answer |= (work << shift);
    shift += 7;
    
    /* do we have the continuation bit on? */
    if ((ch & 0200) == 0) {
      break;
    }
  }
  
  /* advance our pointer */
  *buf_pp = buf_p;
  
  if (val_p != 0L) {
    *val_p = answer;
  }
  return ch_c;
}

/****************************** header routines ******************************/

/*
 * static int read_header_version
 *
 * DESCRIPTION:
 *
 * Read in enough of the header to determine what version we have.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
static	int	read_header_version(diskheap_t *dh_p)
{
  unsigned char	buf[DISKHEAP_MAX_HEADER], *buf_p, *bounds_p;
  char		magic[DISKHEAP_FILE_MAGIC_LEN];
  int		ret;
  
  buf_p = buf;
  bounds_p = buf_p + DISKHEAP_MAX_HEADER;
  
  /* read the 1st block */
  ret = _dh_os_pread(dh_p, buf_p, DISKHEAP_MAX_HEADER,
		     0 /* block-number */, 0 /* offset */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* read the file magic */
  if (get_buf(&buf_p, bounds_p, magic, sizeof(magic)) != sizeof(magic)) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* should match our blurb */
  if (memcmp(magic, DISKHEAP_FILE_MAGIC, DISKHEAP_FILE_MAGIC_LEN) != 0) {
    return DISKHEAP_ERROR_DISK_MAGIC;
  }
  
  /* read the key file type information */
  if (get_uint(&buf_p, bounds_p, &dh_p->dh_version) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * static int read_header_1
 *
 * DESCRIPTION:
 *
 * Read in a header type 1 from a diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
static	int	read_header_1(diskheap_t *dh_p)
{
  unsigned char	buf[DISKHEAP_MAX_HEADER], *buf_p, *bounds_p;
  label_loc_t	*lab_p;
  char		magic[DISKHEAP_FILE_MAGIC_LEN];
  unsigned int	label_n;
  int		ret;
  
  buf_p = buf;
  bounds_p = buf_p + DISKHEAP_MAX_HEADER;
  
  /* read the 1st block */
  ret = _dh_os_pread(dh_p, buf_p, DISKHEAP_MAX_HEADER,
		     0 /* block-number */, 0 /* offset */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* read the file magic */
  if (get_buf(&buf_p, bounds_p, magic, sizeof(magic)) != sizeof(magic)) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* should match our blurb */
  if (memcmp(magic, DISKHEAP_FILE_MAGIC, DISKHEAP_FILE_MAGIC_LEN) != 0) {
    return DISKHEAP_ERROR_DISK_MAGIC;
  }
  
  /* read the key file type information */
  if (get_uint(&buf_p, bounds_p, &dh_p->dh_version) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_block_size) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_user_type) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_min_block_n) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_max_block_n) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* read the dates of various operations */
  if (get_uint(&buf_p, bounds_p, &dh_p->dh_create_time) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_open_time) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_close_time) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_sync_header_time) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_sync_free_time) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* read the stats */
  if (get_uint(&buf_p, bounds_p, &dh_p->dh_alloc_c) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_rewrite_c) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_delete_c) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_max_alloc) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_min_alloc) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_total_blocks) <= 0
      || get_uint(&buf_p, bounds_p, &dh_p->dh_invalid_free_c) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  if (get_uint(&buf_p, bounds_p, &label_n) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  memset(dh_p->dh_labels, 0,
	 sizeof(dh_p->dh_labels[0]) * COUNT_OF(dh_p->dh_labels));
  
  /*
   * the file holds too many labels -- this is bad that the api
   * changed to let this happen
   */
  if (label_n > COUNT_OF(dh_p->dh_labels)) {
    return DISKHEAP_ERROR_LABEL_BOUNDS;
  }
  
  /* read the user labels */
  for (lab_p = dh_p->dh_labels; lab_p < dh_p->dh_labels + label_n; lab_p++) {
    
    /* we store the label string as length,bytes without the \0 */
    if (get_uint(&buf_p, bounds_p, &lab_p->ll_length) <= 0
	/* the string could be 0 bytes long */
	|| get_buf(&buf_p, bounds_p, lab_p->ll_label, lab_p->ll_length) < 0
	|| get_uint(&buf_p, bounds_p, &lab_p->ll_block_num) <= 0
	|| get_uint(&buf_p, bounds_p, &lab_p->ll_offset) <= 0) {
      return DISKHEAP_ERROR_INTERNAL;
    }
  }
  
  /* if the open time is later than the close time, there is problem */
  if (dh_p->dh_open_time > dh_p->dh_close_time) {
    return DISKHEAP_ERROR_BAD_CLOSE;
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * static int write_header_1
 *
 * DESCRIPTION:
 *
 * Assemble and write a header from the diskheap structure.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
static	int	write_header_1(diskheap_t *dh_p)
{
  unsigned char	buf[DISKHEAP_MAX_HEADER], *buf_p, *bounds_p;
  label_loc_t	*lab_p;
  unsigned int	now;
  int		ret;
  
  buf_p = buf;
  bounds_p = buf_p + DISKHEAP_MAX_HEADER;
  
  /* clean the buffer */
  memset(buf_p, 0, DISKHEAP_MAX_HEADER);
  
  /* get the current time */
  now = _dh_os_time();
  
  /* write our file magic blurb */
  if (put_buf(&buf_p, bounds_p, DISKHEAP_FILE_MAGIC,
	      DISKHEAP_FILE_MAGIC_LEN) != DISKHEAP_FILE_MAGIC_LEN) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write the key file type information */
  if (put_uint(&buf_p, bounds_p, dh_p->dh_version) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_block_size) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_user_type) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_min_block_n) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_max_block_n) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write the dates of various operations */
  if (put_uint(&buf_p, bounds_p, dh_p->dh_create_time) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_open_time) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_close_time) <= 0
      /* we put the current time then update the struct below */
      || put_uint(&buf_p, bounds_p, now) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_sync_free_time) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write the stats */
  if (put_uint(&buf_p, bounds_p, dh_p->dh_alloc_c) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_rewrite_c) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_delete_c) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_max_alloc) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_min_alloc) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_total_blocks) <= 0
      || put_uint(&buf_p, bounds_p, dh_p->dh_invalid_free_c) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write the number of labels in the array in case this changes */
  if (put_uint(&buf_p, bounds_p, COUNT_OF(dh_p->dh_labels)) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write the user labels */
  for (lab_p = dh_p->dh_labels;
       lab_p < dh_p->dh_labels + COUNT_OF(dh_p->dh_labels);
       lab_p++) {
    
    /* we store the label string as length,bytes without the \0 */
    if (put_uint(&buf_p, bounds_p, lab_p->ll_length) <= 0
	/* the string could be 0 bytes long */
	|| put_buf(&buf_p, bounds_p, lab_p->ll_label, lab_p->ll_length) < 0
	|| put_uint(&buf_p, bounds_p, lab_p->ll_block_num) <= 0
	|| put_uint(&buf_p, bounds_p, lab_p->ll_offset) <= 0) {
      return DISKHEAP_ERROR_INTERNAL;
    }
  }
  
  /* actually write out the header */
  ret = _dh_os_pwrite(dh_p, buf, DISKHEAP_MAX_HEADER,
		      0 /* block-number */, 0 /* offset */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* note that we just wrote out the header */
  dh_p->dh_sync_header_time = now;
  
  return DISKHEAP_ERROR_NONE;
}

/**************************** free list routines *****************************/

/*
 * static int read_free_1
 *
 * DESCRIPTION:
 *
 * Read in the free list in a version 1 type of file.  The free stuff
 * is stored starting right past the end of the admin header.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
static	int	read_free_1(diskheap_t *dh_p)
{
  unsigned char	buf[DISKHEAP_FREE_CHUNK * 2], *buf_p, *bounds_p;
  unsigned char	*read_p, *start_p;
  unsigned int	where, block_num, size, offset;
  int		ret, left, bytes_b = 0;
  
  /*
   * Read in the free information which is a list of block-number /
   * num-blocks pairs.  The blocks are sorted by size, not location
   * which means that we can't use distance coding.  It was also
   * thought that we did not want to have 1 byte error scrap the
   * entire free list.  An alternative is to keep 2 skip lists one by
   * location and one by size or re-sort the one by location before we
   * write it.
   */
  
  where = DISKHEAP_MAX_HEADER;
  
  /*
   * track were we are reading into because we are shifting around the
   * buffer
   */
  read_p = buf;
  start_p = buf;
  
  while (1) {
    
    /*
     * read a bunch of free stuff which may include the initial user
     * allocations at the front of the file
     */
    ret = _dh_os_pread(dh_p, read_p, DISKHEAP_FREE_CHUNK,
		       0 /* block-number */, where);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    where += DISKHEAP_FREE_CHUNK;
    bounds_p = read_p + DISKHEAP_FREE_CHUNK;
    buf_p = buf;
    
    while (1) {
      
      /*
       * note the start of the entry so we can wind back if we run out
       * of space
       */
      start_p = buf_p;
      
      /*
       * we do this here instead of in the while loop above, because
       * we need to record the start before we check the bounds else
       * we might duplicate the last entry
       */
      if (buf_p >= bounds_p) {
	break;
      }
      
      /* did we run out of space? */
      if (get_uint(&buf_p, bounds_p, &block_num) <= 0) {
	break;
      }
      
      /* are we done? */
      if (block_num == 0) {
	/*
	 * the first 0 means switch to the byte free slots, the second
	 * means we are done
	 */
	if (bytes_b) {
	  /*
	   * yes a goto (gasp) but this is because we need to jump out
	   * of 2 loops.
	   */
	  goto finished;
	}
	else {
	  bytes_b = 1;
	  continue;
	}
      }
      
      /* get the offset if in byte mode */
      if (bytes_b && get_uint(&buf_p, bounds_p, &offset) <= 0) {
	break;
      }
      
      /* get the size in bytes or blocks of the free space */
      if (get_uint(&buf_p, bounds_p, &size) <= 0) {
	break;
      }
      
      /* now add the entry onto the free list */
      if (bytes_b) {
	ret = _dh_free_add_bytes(dh_p, block_num, offset, size);
      }
      else {
	ret = _dh_free_add_blocks(dh_p, block_num, size);
      }
      if (ret != DISKHEAP_ERROR_NONE) {
	return ret;
      }
    }
    
    /*
     * we have a partial entry at the end of the block so we will
     * shift it down and add onto it
     */
    left = bounds_p - start_p;
    memmove(buf, start_p, left);
    read_p = buf + left;
  }
  
  /*
   * yes a goto (gasp) but this is here so we can jump out of the
   * inner loop
   */
 finished:
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * static int write_free_1
 *
 * DESCRIPTION:
 *
 * Write out the free list to the diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - Diskheap error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
static	int	write_free_1(diskheap_t *dh_p)
{
  unsigned char	buf[DISKHEAP_FREE_CHUNK * 2], *buf_p, *bounds_p, *block_end_p;
  unsigned int	where, left, slot_c;
  skip_slot_t	*slot_p;
  sub_slot_t	*sub_slot_p;
  int		ret, bytes_b = 0;
  
  /*
   * We are using an internal buffer here which is 2 times the
   * free-chunk size because we want to be able to add in a free slot
   * which overlaps a chunk without problems.  If an overlap happens,
   * the free-chunk is written out and the excess bytes are shifted
   * down.
   */
  
  where = DISKHEAP_MAX_HEADER;
  
  buf_p = buf;
  bounds_p = buf + sizeof(buf);
  block_end_p = buf + DISKHEAP_FREE_CHUNK;
  
  slot_p = dh_p->dh_skip_block_p->fs_next_p[0];
  slot_c = 0;
  sub_slot_p = dh_p->dh_sub_free_p[0];
  
  /* run through all of the entries dumping them to disk */
  while (1) {
    
    /*
     * if we are processing the free block slots and there are no more
     * then switch to byte free sub-block slots
     */
    if (! bytes_b) {
      
      if (slot_p == NULL) {
	/* add in a 0 which is the boundary marker */
	if (put_uint(&buf_p, bounds_p, 0) <= 0) {
	  return DISKHEAP_ERROR_INTERNAL;
	}
	bytes_b = 1;
      }
      else {
	/* add the pertanent information into the buffer */
	if (put_uint(&buf_p, bounds_p, slot_p->fs_block_num) <= 0
	    || put_uint(&buf_p, bounds_p, slot_p->fs_size) <= 0) {
	  return DISKHEAP_ERROR_INTERNAL;
	}
      }
    }
    /*
     * NOTE: this isn't and if/else here because the above might set
     * the flag so that we go into both ifs
     */
    
    /*
     * if we are processing the free sub-block slots are there are no
     * more in this size linked list then try another slot
     */
    if (bytes_b) {
      if (sub_slot_p == NULL) {
	for (slot_c++; slot_c < COUNT_OF(dh_p->dh_sub_free_p); slot_c++) {
	  sub_slot_p = dh_p->dh_sub_free_p[slot_c];
	  if (sub_slot_p != NULL) {
	    break;
	  }
	}
	if (sub_slot_p == NULL) {
	  break;
	}
      }
      
      /* add the pertanent information into the buffer */
      if (put_uint(&buf_p, bounds_p, sub_slot_p->ss_block_num) <= 0
	  || put_uint(&buf_p, bounds_p, sub_slot_p->ss_offset) <= 0
	  || put_uint(&buf_p, bounds_p, dh_p->dh_sub_sizes[slot_c]) <= 0) {
	return DISKHEAP_ERROR_INTERNAL;
      }
    }
    
    /* do we need to write out the filled block? */
    if (buf_p >= block_end_p) {
      
      /* write a block of free information to disk */
      ret = _dh_os_pwrite(dh_p, buf, DISKHEAP_FREE_CHUNK, 0 /* block-number */,
			  where);
      if (ret != DISKHEAP_ERROR_NONE) {
	return ret;
      }
      where += DISKHEAP_FREE_CHUNK;
      
      if (buf_p == block_end_p) {
	buf_p = buf;
      }
      else {
	/* shift the data down to the front of the block */
	left = buf_p - block_end_p;
	memmove(buf, block_end_p, left);
	buf_p = buf + left;
      }
    }

    if (bytes_b) {
      sub_slot_p = sub_slot_p->ss_next_p;
    }
    else {
      slot_p = slot_p->fs_next_p[0];
    }
  }
  
  /* put three 0s at the end of the block */
  if (put_uint(&buf_p, bounds_p, 0) <= 0
      || put_uint(&buf_p, bounds_p, 0) <= 0
      || put_uint(&buf_p, bounds_p, 0) <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* write the last block(s) */
  while (buf_p > buf) {
    
    if (buf_p < block_end_p) {
      /* clear out the end of the block */
      memset(buf_p, 0, block_end_p - buf_p);
    }
    
    /* write a block of free information to disk */
    ret = _dh_os_pwrite(dh_p, buf, DISKHEAP_FREE_CHUNK, 0 /* block-number */,
			where);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    where += DISKHEAP_FREE_CHUNK;
    
    if (buf_p <= block_end_p) {
      break;
    }
    
    /* shift the data down to the front of the block */
    left = buf_p - block_end_p;
    if (left > 0) {
      memmove(buf, block_end_p, left);
      buf_p = buf + left;
    }
  }
  
  return DISKHEAP_ERROR_NONE;
}

/***************************** exported routines *****************************/

/*
 * int _dh_read_header
 *
 * DESCRIPTION:
 *
 * Read the disk header from a diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
int	_dh_read_header(diskheap_t *dh_p)
{
  int	ret;
  
  /* if the version number is unknown then set it */
  if (dh_p->dh_version == 0) {
    ret = read_header_version(dh_p);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
  }
  
  switch (dh_p->dh_version) {
    
  case 1:
    return read_header_1(dh_p);
    break;
  }
  
  return DISKHEAP_ERROR_VERSION;
}

/*
 * int _dh_write_header
 *
 * DESCRIPTION:
 *
 * Write the header into a disk heap.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
int	_dh_write_header(diskheap_t *dh_p)
{
  switch (dh_p->dh_version) {
    
  case 1:
    return write_header_1(dh_p);
    break;
  }
  
  return DISKHEAP_ERROR_VERSION;
}

/*
 * int _dh_read_free
 *
 * DESCRIPTION:
 *
 * Read the free list from a diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
int	_dh_read_free(diskheap_t *dh_p)
{
  switch (dh_p->dh_version) {
    
  case 1:
    return read_free_1(dh_p);
    break;
  }
  
  return DISKHEAP_ERROR_VERSION;
}


/*
 * int _dh_write_free
 *
 * DESCRIPTION:
 *
 * Write out the free list to the diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
int	_dh_write_free(diskheap_t *dh_p)
{
  switch (dh_p->dh_version) {
    
  case 1:
    return write_free_1(dh_p);
    break;
  }
  
  return DISKHEAP_ERROR_VERSION;
}

/*
 * int _dh_write_initial_user
 *
 * DESCRIPTION:
 *
 * Write a block of 0s at the start of the user allocation section.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
int	_dh_write_initial_user(diskheap_t *dh_p)
{
  /* clean the buffer */
  memset(dh_p->dh_buf, 0, dh_p->dh_block_size);
  
  /* write a block of 0s at the start of the user space */
  return _dh_os_pwrite(dh_p, dh_p->dh_buf, dh_p->dh_block_size,
		       dh_p->dh_min_block_n, 0 /* offset */);
}

/*
 * int _dh_block_header
 *
 * DESCRIPTION:
 *
 * Write the block header information into the start of the internal
 * block buffer.
 *
 * RETURNS:
 *
 * Success - Number of bytes written into the header buffer.
 *
 * Failure - -1
 *
 * ARGUMENTS:
 *
 * buf <-> Character buffer into which we will be writing the header.
 *
 * bounds_p <-> Pointer past the end of the character buffer.
 *
 * head_p -> Pointer to the header information we are storing in the
 * block.
 */
int	_dh_block_header(unsigned char *buf, const unsigned char *bounds_p,
			 const block_head_t *head_p)
{
  unsigned char	*buf_p;
  
  /*
   * Block header: minimum 5 + user-data, maximum 17 + user-data
   *
   * MMBUSFNDDD...
   *
   * MM = 2 bytes of block magic number (all)
   *
   * B - block-type (all)
   *
   * U - user specified type (user-all, user-start)
   *
   * S - user requested size in bytes (user-all, user-start)
   *
   * F - block-number of first block in chain (user-middle, user-end)
   *
   * N - block-number of next block in chain (user-start, user-middle)
   *
   * DDD... - user data
   */
  
  buf_p = buf;
  
  /* put in the block magic and type */
  if (put_uint(&buf_p, bounds_p, BLOCK_MAGIC_NUMBER) <= 0
      || put_uint(&buf_p, bounds_p, head_p->bh_block_type) <= 0) {
    return -1;
  }
  
  /* add in the sub-block size in bytes */
  if (head_p->bh_block_type == BLOCK_TYPE_SUB
      && put_uint(&buf_p, bounds_p, head_p->bh_user_size) <= 0) {
    return -1;
  }
  
  /*
   * start of user allocations and free block gets the total number of
   * blocks in the block
   */
  if (head_p->bh_block_type == BLOCK_TYPE_USER_ALL
      || head_p->bh_block_type == BLOCK_TYPE_FREE) {
    if (put_uint(&buf_p, bounds_p, head_p->bh_block_n) <= 0) {
      return -1;
    }
  }
  
  /*
   * if we have an entire user allocation or the start of one, write
   * in the user specified type and the real size of the allocation in
   * bytes
   */
  if (head_p->bh_block_type == BLOCK_TYPE_USER_ALL) {
    if (put_uint(&buf_p, bounds_p, head_p->bh_user_type) <= 0
	|| put_uint(&buf_p, bounds_p, head_p->bh_user_size) <= 0) {
      return -1;
    }
  }
  
  return buf_p - buf;
}

/*
 * int _dh_sub_block_header
 *
 * DESCRIPTION:
 *
 * Write the sub-block header information into the start of the
 * internal block buffer.
 *
 * RETURNS:
 *
 * Success - Number of bytes written into the header buffer.
 *
 * Failure - -1
 *
 * ARGUMENTS:
 *
 * buf <-> Character buffer into which we will be writing the header.
 *
 * bounds_p <-> Pointer past the end of the character buffer.
 *
 * head_p -> Pointer to the header information we are storing in the
 * sub-block.
 */
int	_dh_sub_block_header(unsigned char *buf, const unsigned char *bounds_p,
			     const block_head_t *head_p)
{
  unsigned char	*buf_p;
  
  buf_p = buf;
  
  /* put the type of the block: either USER_ALL or FREE */
  if (put_uint(&buf_p, bounds_p, head_p->bh_block_type) <= 0) {
    return -1;
  }
  
  /*
   * if we have an entire user allocation, write in the user specified
   * type and the real size of the allocation in bytes
   */
  if (head_p->bh_block_type == BLOCK_TYPE_USER_ALL
      && (put_uint(&buf_p, bounds_p, head_p->bh_user_type) <= 0
	  || put_uint(&buf_p, bounds_p, head_p->bh_user_size) <= 0)) {
    return -1;
  }
  
  return buf_p - buf;
}

/*
 * int _dh_write_header_user
 *
 * DESCRIPTION:
 *
 * Write the header in the dh buffer and the user area to block(s).
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 *
 * head_len -> Length in the header in bytes in the dh_p block buffer.
 *
 * buffer -> User's anonymous memory to be written to disk.
 *
 * user_size -> Size of the user memory.  Set to 0 to just write out a
 * new header.
 *
 * block_num -> Block-number were we are writing it.  No offset
 * necessary since we are writing a block header here.
 */
int	_dh_write_header_user(diskheap_t *dh_p, const int head_len,
			      const unsigned char *buffer,
			      const unsigned int user_size,
			      const unsigned int block_num)
{
  unsigned char	*buf_p;
  unsigned int	where, left;
  int		ret;
  
  left = dh_p->dh_block_size - head_len;
  if (left > user_size) {
    left = user_size;
  }
  
  buf_p = dh_p->dh_buf + head_len;
  memmove(buf_p, buffer, left);
  buf_p += left;
  
  /* 0 out the rest of the block if we didn't use it all up */
  if (buf_p < dh_p->dh_buf_bounds_p) {
    memset(buf_p, 0, dh_p->dh_buf_bounds_p - buf_p);
  }
  
  /* write out the 1st block */
  where = block_num;
  ret = _dh_os_pwrite(dh_p, dh_p->dh_buf, dh_p->dh_block_size, where,
		      0 /* offset */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  where++;
  
  /* is there more to write? */
  if (left < user_size) {
    /* write the rest of the user space */
    ret = _dh_os_pwrite(dh_p, buffer + left, user_size - left, where,
			0 /* offset */);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_read_first_block
 *
 * DESCRIPTION:
 *
 * Read in a block header plus a certain number of bytes of the user
 * space up to a block-size worth into the internal buffer space.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 *
 * block_num -> Block-number were we are reading from.
 *
 * max_read_size -> Set the maximum number of bytes to read from the
 * data item or 0 to read all of the bytes.
 *
 * head_p <-> Pointer to a header structure which will be set with the
 * details about the block.
 *
 * head_len_p <- Pointer to an integer which, if not NULL, will be set
 * to the length of the header at the front of the block.
 */
int	_dh_read_first_block(diskheap_t *dh_p, const unsigned int block_num,
			     const unsigned int max_read_size,
			     block_head_t *head_p, int *head_len_p)
{
  unsigned int	magic, read_size;
  unsigned char	*buf_p, *bounds_p;
  int		ret;
  
  if (max_read_size == 0
      || max_read_size > dh_p->dh_block_size - BLOCK_MAX_HEAD_LEN) {
    read_size = dh_p->dh_block_size;
  }
  else {
    read_size = BLOCK_MAX_HEAD_LEN + max_read_size;
  }
  
  /* read in the 1st block */
  ret = _dh_os_pread(dh_p, dh_p->dh_buf, read_size, block_num, 0 /* offset */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /*
   * Block header: minimum 5 + user-data, maximum 17 + user-data
   *
   * MMBUSFNDDD...
   *
   * MM = 2 bytes of block magic number (all)
   *
   * B - block-type (all)
   *
   * U - 1-5 bytes of user specified type (user-all, user-start)
   *
   * S - 1-5 bytes user requested size in bytes (user-all, user-start)
   *
   * F - 1-5 bytes block-number of first block in chain (user-middle,
   * user-end)
   *
   * N - 1-5 bytes block-number of next block in chain (user-start,
   * user-middle)
   *
   * DDD... - user data
   */
  
  buf_p = dh_p->dh_buf;
  bounds_p = dh_p->dh_buf + read_size;
  
  /* get the block magic number */
  if (get_uint(&buf_p, bounds_p, &magic) <= 0
      || magic != BLOCK_MAGIC_NUMBER) {
    return DISKHEAP_ERROR_BLOCK_HEADER;
  }
  
  /* clear out the structure */
  memset(head_p, 0, sizeof(*head_p));
  
  /* get the block byte */
  if (get_uint(&buf_p, bounds_p, &head_p->bh_block_type) <= 0) {
    return DISKHEAP_ERROR_BLOCK_HEADER;
  }
  
  /* if this is a sub-block, get the slot-size in bytes */
  if (head_p->bh_block_type == BLOCK_TYPE_SUB
      && get_uint(&buf_p, bounds_p, &head_p->bh_user_size) <= 0) {
    return -1;
  }
  
  /*
   * every block gets the total number of blocks except the
   * continuation sections
   */
  if (head_p->bh_block_type == BLOCK_TYPE_USER_ALL
      || head_p->bh_block_type == BLOCK_TYPE_FREE) {
    if (get_uint(&buf_p, bounds_p, &head_p->bh_block_n) <= 0) {
      return DISKHEAP_ERROR_BLOCK_HEADER;
    }
  }
  
  /*
   * if we have an entire user allocation or the start of one, write
   * in the user specified type and the real size of the allocation in
   * bytes
   */
  if (head_p->bh_block_type == BLOCK_TYPE_USER_ALL) {
    if (get_uint(&buf_p, bounds_p, &head_p->bh_user_type) <= 0
	|| get_uint(&buf_p, bounds_p, &head_p->bh_user_size) <= 0) {
      return DISKHEAP_ERROR_BLOCK_HEADER;
    }
  }
  
  SET_POINTER(head_len_p, buf_p - dh_p->dh_buf);
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_read_sub_block
 *
 * DESCRIPTION:
 *
 * Read in a sub-block.  We first read in and verify the front of the
 * block then we verify and move down to the offset in question.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 *
 * block_num -> Block-number were we are reading from.
 *
 * offset -> Offset where we are writing.
 *
 * buffer -> Buffer to store the data in or NULL to alloc.
 *
 * max_read_size -> Set the maximum number of bytes to read from the
 * data item or 0 to read all of the bytes.
 *
 * head_p <-> Pointer to a header structure which will be set with the
 * details about the block.
 *
 * slot_size_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the number of bytes in the sub-block slot-size.
 *
 * user_size_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the number of bytes in the user's on-disk chunk, not
 * including any internal header information.
 *
 * user_type_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the type value which was passed to the store
 * function.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
void	*_dh_read_sub_block(diskheap_t *dh_p, const unsigned int block_num,
			    const unsigned int offset,
			    void *buffer, const unsigned int max_read_size,
			    unsigned int *slot_size_p,
			    unsigned int *user_size_p,
			    unsigned int *user_type_p, int *error_p)
{
  unsigned int	read_size, slot_size, user_size, buf_size;
  unsigned int	head_len, sub_type, size_c;
  unsigned char	*buf_p, *bounds_p, *user_buf;
  block_head_t	head;
  int		ret;
  
  /* read in the block header for the sub-block */
  ret = _dh_read_first_block(dh_p, block_num, 1 /* read minimal user bytes */,
			     &head, &head_len);
  if (ret != DISKHEAP_ERROR_NONE) {
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  /* we better be a sub-block */
  if (head.bh_block_type != BLOCK_TYPE_SUB) {
    SET_POINTER(error_p, DISKHEAP_ERROR_LOCATION);
    return NULL;
  }
  slot_size = head.bh_user_size;
  
  /*
   * the caller may want to know this if they are free-ing after
   * reading it to verify it
   */
  SET_POINTER(slot_size_p, slot_size);
  
  /* make sure our slot size > 0 */
  if (slot_size == 0) {
    SET_POINTER(error_p, DISKHEAP_ERROR_BLOCK_HEADER);
    return NULL;
  }
  
  /* verify our slot size matches an entry in our array */
  for (size_c = 0; size_c < COUNT_OF(dh_p->dh_sub_sizes); size_c++) {
    if (slot_size == dh_p->dh_sub_sizes[size_c]) {
      break;
    }
  }
  
  /* error if none matched */
  if (size_c >= COUNT_OF(dh_p->dh_sub_sizes)) {
    SET_POINTER(error_p, DISKHEAP_ERROR_BLOCK_HEADER);
    return NULL;
  }
  
  /* verify that our offset is not out of bounds and is aligned incorrectly */
  if (offset > dh_p->dh_block_size - slot_size
      || (offset - head_len) % slot_size != 0) {
    SET_POINTER(error_p, DISKHEAP_ERROR_LOCATION);
    return NULL;
  }
  
  /* see how much of the sub-block we should read in */
  if (max_read_size == 0
      || BLOCK_MAX_HEAD_LEN + max_read_size > slot_size) {
    read_size = slot_size;
  }
  else {
    read_size = BLOCK_MAX_HEAD_LEN + max_read_size;
  }
  
  /* now read in the sub-block allocation */
  ret = _dh_os_pread(dh_p, dh_p->dh_buf, read_size, block_num, offset);
  if (ret != DISKHEAP_ERROR_NONE) {
    SET_POINTER(error_p, ret);
    return NULL;
  }
  
  bounds_p = dh_p->dh_buf + read_size;
  buf_p = dh_p->dh_buf;
  
  /* get the type from the sub-block header */
  if (get_uint(&buf_p, bounds_p, &sub_type) <= 0) {
    SET_POINTER(error_p, DISKHEAP_ERROR_INTERNAL);
    return NULL;
  }
  
  /* verify that it is allocated */
  if (sub_type != BLOCK_TYPE_USER_ALL) {
    if (sub_type == BLOCK_TYPE_FREE) {
      SET_POINTER(error_p, DISKHEAP_ERROR_BLOCK_FREE);
    }
    else {
      SET_POINTER(error_p, DISKHEAP_ERROR_LOCATION);
    }
    return NULL;
  }
  
  /*
   * if we have an entire user allocation, write in the user specified
   * type and the real size of the allocation in bytes
   */
  if (get_uint(&buf_p, bounds_p, user_type_p) <= 0
      || get_uint(&buf_p, bounds_p, &user_size) <= 0) {
    SET_POINTER(error_p, DISKHEAP_ERROR_INTERNAL);
    return NULL;
  }
  
  /* how much are we actually returning in the user buffer */
  if (max_read_size == 0 || max_read_size >= user_size) {
    buf_size = user_size;
  }
  else {
    buf_size = max_read_size;
  }
  
  /* problem if we don't have enough left */
  if (buf_p + buf_size > bounds_p) {
    SET_POINTER(error_p, DISKHEAP_ERROR_INTERNAL);
    return NULL;
  }
  
  if (buffer == NULL) {
    user_buf = malloc(buf_size);
    if (user_buf == NULL) {
      SET_POINTER(error_p, DISKHEAP_ERROR_ALLOC);
      return NULL;
    }
  }
  else {
    user_buf = buffer;
  }
  
  /* copy the user allocation into the user buf */
  memmove(user_buf, buf_p, buf_size);
  
  SET_POINTER(user_size_p, user_size);
  SET_POINTER(error_p, DISKHEAP_ERROR_NONE);
  return user_buf;
}
