/*
 * Diskheap Library Sub-Block Routines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_sub.c,v 1.6 2002/02/19 15:32:01 gray Exp $
 */

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define __DISKHEAP_MAIN__

#include "conf.h"
#include "compat.h"
#include "diskheap.h"
#include "diskheap_free.h"
#include "diskheap_io.h"
#include "diskheap_os.h"
#include "diskheap_loc.h"
#include "diskheap_sub.h"

#if IDENT_WORKS
#ident "$Id: diskheap_sub.c,v 1.6 2002/02/19 15:32:01 gray Exp $"
#else
static char *rcs_id = "$Id: diskheap_sub.c,v 1.6 2002/02/19 15:32:01 gray Exp $";
#endif

/*************************** accounting functions ****************************/

/*
 * int _dh_account_alloc
 *
 * DESCRIPTION:
 *
 * Adjust the stats to account for an allocation.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to the diskheap we are working on.
 *
 * block_n -> How many blocks we are using.  0 if none.
 *
 * size -> Size of the user allocation.
 *
 * end_b -> Set to 1 if the blocks were allocated at the end of the
 * heap.
 */
int	_dh_account_alloc(diskheap_t *dh_p, const unsigned int block_n,
			 const unsigned int size, const int end_b)
{
  int	ret;
  
  dh_p->dh_alloc_c++;
  if (size > dh_p->dh_max_alloc) {
    dh_p->dh_max_alloc = size;
  }
  if (size < dh_p->dh_min_alloc || dh_p->dh_min_alloc == 0) {
    dh_p->dh_min_alloc = size;
  }
  dh_p->dh_total_blocks += block_n;
  
  if (end_b) {
    /* adjust our max block size upwards */
    dh_p->dh_max_block_n += block_n;
  }
  else {
    /* see if we need to auto-sync the free list to disk */
    if (dh_p->dh_sync_free > 0) {
      dh_p->dh_sync_free_list_c++;
      if (dh_p->dh_sync_free_list_c >= dh_p->dh_sync_free) {
	ret = _dh_write_free(dh_p);
	if (ret != DISKHEAP_ERROR_NONE) {
	  return ret;
	}
	dh_p->dh_sync_free_list_c = 0;
      }
    }
  }
  
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_account_delete
 *
 * DESCRIPTION:
 *
 * Adjust the stats to account for a deletion.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to the diskheap we are working on.
 *
 * block_n -> How many blocks we are deleting.  0 if none.
 */
int	_dh_account_delete(diskheap_t *dh_p, const unsigned int block_n)
{
  int	ret;
  
  dh_p->dh_delete_c++;
  dh_p->dh_total_blocks -= block_n;
  
  /* see if we need to auto-sync the free list to disk */
  if (dh_p->dh_sync_free > 0) {
    dh_p->dh_sync_free_list_c++;
    if (dh_p->dh_sync_free_list_c >= dh_p->dh_sync_free) {
      ret = _dh_write_free(dh_p);
      if (ret != DISKHEAP_ERROR_NONE) {
	return ret;
      }
      dh_p->dh_sync_free_list_c = 0;
    }
  }
  
  return DISKHEAP_ERROR_NONE;
}

/************************ sub-block exported routines ************************/

/*
 * int _dh_sub_init
 *
 * DESCRIPTION:
 *
 * Initialize the sub-block allocation routines.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are initializing.
 */
int	_dh_sub_init(diskheap_t *dh_p)
{
  int		divisor, min_head_len, head_len, new_head_len;
  unsigned int	size_c, largest_slot = 0;
  block_head_t	head;
  
  memset(&head, 0, sizeof(head));
  head.bh_block_type = BLOCK_TYPE_SUB;
  
  /* get the minimum header size */
  min_head_len = _dh_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p, &head);
  if (min_head_len <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* sanity check here */
  if (COUNT_OF(dh_p->dh_sub_sizes) != COUNT_OF(dh_p->dh_sub_free_p)) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /*
   * figure out the largest divisor which results in at least the
   * minimum sub-block size
   */
  for (divisor = 2;
       dh_p->dh_block_size / divisor >= MINIMUM_SUB_BLOCK_SIZE;
       divisor *= 2) {
  }
  divisor /= 2;
  
  /*
   * We iterate here to calculate the try sizes of each of the
   * sub-blocks.  Since we have block header information, we can't
   * just take the block-size / 2,4,8, etc. but instead we have to
   * take the block-size - block-header and divide.  The added
   * challenge is that we don't know the size of the block-header
   * until we no how big the user chunks are, hence the while loop
   * inside.
   */
  for (size_c = 0; size_c < COUNT_OF(dh_p->dh_sub_sizes); size_c++) {
    
    new_head_len = min_head_len;
    
    /*
     * we iterate here because we don't know how big the header is
     * going to be until we get the size
     */
    do {
      head_len = new_head_len;
      head.bh_user_size = (dh_p->dh_block_size - head_len) / divisor;
      new_head_len = _dh_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p,
				      &head);
      if (new_head_len <= 0) {
	return DISKHEAP_ERROR_INTERNAL;
      }
    } while (new_head_len > head_len);
    
    dh_p->dh_sub_sizes[size_c] = head.bh_user_size;
    /* this only gets bigger */
    largest_slot = head.bh_user_size;
    divisor /= 2;
    if (divisor <= 1) {
      break;
    }
  }
  
  /*
   * record the largest size that goes in a sub-block which includes a
   * basic header
   */
  memset(&head, 0, sizeof(head));
  head.bh_block_type = BLOCK_TYPE_USER_ALL;
  head.bh_user_type = 0;
  /*
   * we iterate here because we don't know how big the user-header is
   * going to be until we get its size and try to encode it
   */
  dh_p->dh_sub_largest_size = dh_p->dh_block_size;
  do {
    head.bh_user_size = dh_p->dh_sub_largest_size;
    head_len = _dh_sub_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p,
				    &head);
    if (head_len <= 0) {
      return DISKHEAP_ERROR_INTERNAL;
    }
    dh_p->dh_sub_largest_size = largest_slot - head_len;
  } while (dh_p->dh_sub_largest_size < head.bh_user_size);
  
  /* calculate our size array and the largest_size global */
  return DISKHEAP_ERROR_NONE;
}

/*
 * int _dh_sub_store
 *
 * DESCRIPTION:
 *
 * Store a sub-block sized piece of memory.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * buffer -> Buffer of anonymous memory to be written to disk.
 *
 * user_size -> Size of buffer in bytes.  Must be > 0.
 *
 * user_type -> Type number to be associated with the buffer that is
 * being stored.  This number will be returned when you retrieve the
 * block and can be used to verify that you are dealing with the right
 * kind of storage chunk.  Pass in 0 if none is desired.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number of the allocation.  Both the block-number and the offset are
 * needed to read the information later.  This argument cannot be
 * NULL.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the allocation.  Both the block-number and
 * the offset are needed to read the information later.  This argument
 * cannot be NULL.
 */
int	_dh_sub_store(diskheap_t *dh_p, const void *buffer,
		      const unsigned int user_size,
		      const unsigned int user_type,
		      unsigned int *block_num_p, unsigned int *offset_p)
{
  int		head_len, sub_head_len, ret, found_n, end_b = 0;
  unsigned int	needed, block_num, offset, slot_size;
  unsigned char	*buf_p, *bounds_p, head_buf[BLOCK_MAX_HEAD_LEN];
  block_head_t	head;
  
  /* setup our sub-block header */
  memset(&head, 0, sizeof(head));
  head.bh_block_type = BLOCK_TYPE_USER_ALL;
  head.bh_user_type = user_type;
  head.bh_user_size = user_size;
  
  /* figure out how big the final block needs to be */
  sub_head_len = _dh_sub_block_header(head_buf, head_buf + sizeof(head_buf),
				      &head);
  if (sub_head_len <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* now find out exactly how much space we need */
  needed = user_size + sub_head_len;
  if (needed > dh_p->dh_sub_largest_size) {
    return DISKHEAP_ERROR_TOO_LARGE;
  }
  if (needed > dh_p->dh_block_size) {
    /* what??  this shouldn't happen */
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /*
   * We loop here until we get a free entry.  We may get a free slot
   * from the free list which really isn't free.
   */
  while (1) {
    char	user_buf[1];
    
    /* check the free-list for the slot in question */
    ret = _dh_free_get_bytes(dh_p, needed, &block_num, &offset, &slot_size);
    if (ret != DISKHEAP_ERROR_NONE) {
      break;
    }
    
    /* verify that we have a free slot here */
    if (_dh_read_sub_block(dh_p, block_num, offset, user_buf,
			   1 /* read minimum user bytes */,
			   NULL /* slot-size */, NULL /* user-size */,
			   NULL /* user-type */, &ret) != NULL
	|| ret != DISKHEAP_ERROR_BLOCK_FREE) {
      /* the bad entry has already been removed and we go on */
      dh_p->dh_invalid_free_c++;
      continue;
    }
    
    /* copy the header into the block buffer */
    memmove(dh_p->dh_buf, head_buf, sub_head_len);
    
    /* copy the user bytes into our internal buffer */
    memmove(dh_p->dh_buf + sub_head_len, buffer, user_size);
    
    /* write the sub-block to disk */
    ret = _dh_os_pwrite(dh_p, dh_p->dh_buf, needed, block_num, offset);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    SET_POINTER(block_num_p, block_num);
    SET_POINTER(offset_p, offset);
    
    /* adjust our stats */
    return _dh_account_alloc(dh_p, 0 /* no blocks used */, user_size,
			     0 /* no alloc at end of heap */);
  }
  
  /* if there was some other error, then return it */
  if (ret != DISKHEAP_ERROR_NO_FREE) {
    return ret;
  }
  
  /*
   * Since we don't have a sub-block of the appropriate size, we have
   * to allocate a block to hold the sub-blocks of this size.
   */
  
  ret = _dh_free_get_one(dh_p, &block_num, &found_n);
  if (ret == DISKHEAP_ERROR_NONE) {
    /*
     * all we need is 1 block so we might as well put all but the
     * first block back on the free list
     */
    if (found_n > 1) {
      /* write the new free block header */
      head.bh_block_type = BLOCK_TYPE_FREE;
      head.bh_block_n = found_n - 1;
      
      /* keep the same block-size and other user information */
      head_len = _dh_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p, &head);
      if (head_len < 0) {
	return DISKHEAP_ERROR_INTERNAL;
      }
      
      /* write out the header and no user data into the first block */
      ret = _dh_write_header_user(dh_p, head_len, NULL /* user buffer */,
				  0 /* user-size */, block_num + 1);
      if (ret != DISKHEAP_ERROR_NONE) {
	return ret;
      }
      
      /* now add it to the free list */
      ret = _dh_free_add_blocks(dh_p, block_num + 1, found_n - 1);
      if (ret != DISKHEAP_ERROR_NONE) {
	return ret;
      }
    }
  }
  else if (ret == DISKHEAP_ERROR_NO_FREE) {
    /*
     * need to get a new block of these sized chunks, write out the
     * header, and add the other blocks to the free list
     */
    /* allocate space onto the end of the file */
    end_b = 1;
    block_num = dh_p->dh_max_block_n;
    found_n = 1;
  }
  else if (ret != DISKHEAP_ERROR_NONE) {
    /* if there was some other error, then return it */
    return ret;
  }
  
  /*
   * write new block header into the front of our internal buffer with
   * the bounds being the start of the sub-block header we created
   * above
   */
  
  /* setup our block header for the sub-blocks */
  memset(&head, 0, sizeof(head));
  head.bh_block_type = BLOCK_TYPE_SUB;
  head.bh_user_size = slot_size;
  
  /* build a block header */
  head_len = _dh_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p, &head);
  if (head_len <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* shift the sub-block down next to the block header */
  buf_p = dh_p->dh_buf + head_len;
  memmove(buf_p, head_buf, sub_head_len);
  buf_p += sub_head_len;
  
  /* copy the user bytes into our internal buffer */
  memmove(buf_p, buffer, user_size);
  buf_p += user_size;
  
  needed += head_len;
  /* zero the rest of the buffer */
  memset(buf_p, 0, dh_p->dh_block_size - needed);
  
  /* now we are creating the sub-block headers */
  memset(&head, 0, sizeof(head));
  head.bh_block_type = BLOCK_TYPE_FREE;
  
  /* now we need to add in free headers at the other spots in the buffer */
  offset = head_len + slot_size;
  bounds_p = dh_p->dh_buf + dh_p->dh_block_size;
  for (buf_p = dh_p->dh_buf + offset;
       buf_p <= bounds_p - slot_size;
       buf_p += slot_size) {
    
    /* create the sub-block header */
    if (_dh_sub_block_header(buf_p, bounds_p, &head) <= 0) {
      return DISKHEAP_ERROR_INTERNAL;
    }
    
    /* add it to the free list */
    ret = _dh_free_add_bytes(dh_p, block_num, offset, slot_size);
    if (ret != DISKHEAP_ERROR_NONE) {
      return ret;
    }
    
    offset += slot_size;
  }
  
  /*
   * write the complete block with the user allocation in the
   * sub-block slot #1 and free sub-blocks after it
   */
  ret = _dh_os_pwrite(dh_p, dh_p->dh_buf, dh_p->dh_block_size, block_num,
		      0 /* offset */);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  SET_POINTER(block_num_p, block_num);
  /*
   * the offset is the header-length since the first sub-block
   * allocation stats right after the block-header
   */
  SET_POINTER(offset_p, head_len);
  return _dh_account_alloc(dh_p, 1 /* 1 block used */, user_size, end_b);
}

/*
 * int _dh_sub_delete
 *
 * DESCRIPTION:
 *
 * Delete a previously stored sub-block allocation.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are deleting.
 *
 * offset -> Offset inside of the block of the allocation we are
 * deleting.
 *
 * user_size_p <- Size of the user allocation in bytes.
 *
 * user_type_p <- Type number associated with the buffer by
 * diskheap_store.
 */
int	_dh_sub_delete(diskheap_t *dh_p, const unsigned int block_num,
		       const unsigned int offset,
		       unsigned int *user_size_p,
		       unsigned int *user_type_p)
{
  unsigned int	slot_size;
  block_head_t	head;
  int		ret, head_len;
  
  /* try to read in the sub-block */
  if (_dh_read_sub_block(dh_p, block_num, offset, dh_p->dh_buf,
			 1 /* read minimal user bytes */, &slot_size,
			 user_size_p, user_type_p, &ret) == NULL) {
    return ret;
  }
  
  /* re-write it as a free block */
  memset(&head, 0, sizeof(head));  
  head.bh_block_type = BLOCK_TYPE_FREE;
  
  /* create our header */
  head_len = _dh_sub_block_header(dh_p->dh_buf, dh_p->dh_buf_bounds_p, &head);
  if (head_len <= 0) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* sanity check */
  if ((unsigned int)head_len > slot_size) {
    return DISKHEAP_ERROR_INTERNAL;
  }
  
  /* clear the rest of the sub-block */
  memset(dh_p->dh_buf + head_len, 0, slot_size - head_len);
  
  /* write the sub-block to disk */
  ret = _dh_os_pwrite(dh_p, dh_p->dh_buf, slot_size, block_num, offset);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* add its bytes onto our free list */
  ret = _dh_free_add_bytes(dh_p, block_num, offset, slot_size);
  if (ret != DISKHEAP_ERROR_NONE) {
    return ret;
  }
  
  /* adjust our stats */
  return _dh_account_delete(dh_p, 0);
}
